"""
These solvers solve the (pure) monodomain equations on the form: find
the transmembrane potential :math:`v = v(x, t)` such that

.. math::

   v_t - \mathrm{div} ( G_i v) = I_s

where the subscript :math:`t` denotes the time derivative; :math:`G_i`
denotes a weighted gradient: :math:`G_i = M_i \mathrm{grad}(v)` for,
where :math:`M_i` is the intracellular cardiac conductivity tensor;
:math:`I_s` ise prescribed input. In addition, initial conditions are
given for :math:`v`:

.. math::

   v(x, 0) = v_0

Finally, boundary conditions must be prescribed. For now, this solver
assumes pure homogeneous Neumann boundary conditions for :math:`v`.

"""

# Copyright (C) 2013 Johan Hake (hake@simula.no)
# Use and modify at will
# Last changed: 2013-04-18

__all__ = ["BasicMonodomainSolver", "MonodomainSolver"]
import ufl
import dolfin
from cbcbeat.dolfinimport import backend
from cbcbeat.markerwisefield import rhs_with_markerwise_field
from cbcbeat.utils import end_of_time, annotate_kwargs
from ufl.log import debug, info, error


class BasicMonodomainSolver:
    """This solver is based on a theta-scheme discretization in time
    and CG_1 elements in space.

    .. note::

       For the sake of simplicity and consistency with other solver
       objects, this solver operates on its solution fields (as state
       variables) directly internally. More precisely, solve (and
       step) calls will act by updating the internal solution
       fields. It implies that initial conditions can be set (and are
       intended to be set) by modifying the solution fields prior to
       simulation.

    *Arguments*
      mesh (:py:class:`dolfin.Mesh`)
        The spatial domain (mesh)

      time (:py:class:`dolfin.Constant` or None)
        A constant holding the current time. If None is given, time is
        created for you, initialized to zero.

      M_i (:py:class:`ufl.Expr`)
        The intracellular conductivity tensor (as an UFL expression)

      I_s (:py:class:`dict`, optional)
        A typically time-dependent external stimulus given as a dict,
        with domain markers as the key and a
        :py:class:`dolfin.Expression` as values. NB: it is assumed
        that the time dependence of I_s is encoded via the 'time'
        Constant.

      v\_ (:py:class:`ufl.Expr`, optional)
        Initial condition for v. A new :py:class:`dolfin.Function`
        will be created if none is given.

      params (:py:class:`dolfin.Parameters`, optional)
        Solver parameters

    """

    def __init__(self, mesh, time, M_i, I_s=None, v_=None, params=None):
        # Check some input
        assert isinstance(mesh, dolfin.Mesh), (
            "Expecting mesh to be a Mesh instance, not %r" % mesh
        )
        assert (
            isinstance(time, backend.Constant) or time is None
        ), "Expecting time to be a Constant instance (or None)."
        assert (
            isinstance(params, dolfin.Parameters) or params is None
        ), "Expecting params to be a Parameters instance (or None)"

        # Store input
        self._mesh = mesh
        self._M_i = M_i
        self._I_s = I_s
        self._time = time

        # Initialize and update parameters if given
        self.parameters = self.default_parameters()
        if params is not None:
            self.parameters.update(params)

        # Set-up function spaces
        k = self.parameters["polynomial_degree"]
        family = self.parameters["family"]

        element = dolfin.FiniteElement(
            family=family,
            cell=self._mesh.ufl_cell(),
            degree=k,
            quad_scheme="default",
        )
        V = dolfin.FunctionSpace(self._mesh, element)

        self.V = V

        # Set-up solution fields:
        if v_ is None:
            self.v_ = backend.Function(V, name="v_")
        else:
            debug("Experimental: v_ shipped from elsewhere.")
            self.v_ = v_

        self.v = backend.Function(self.V, name="v")

        # Figure out whether we should annotate or not
        self._annotate_kwargs = annotate_kwargs(self.parameters)

    @property
    def time(self):
        "The internal time of the solver."
        return self._time

    def solution_fields(self):
        """
        Return tuple of previous and current solution objects.

        Modifying these will modify the solution objects of the solver
        and thus provides a way for setting initial conditions for
        instance.

        *Returns*
          (previous v, current v) (:py:class:`tuple` of :py:class:`dolfin.Function`)
        """
        return (self.v_, self.v)

    def solve(self, interval, dt=None):
        """
        Solve the discretization on a given time interval (t0, t1)
        with a given timestep dt and return generator for a tuple of
        the interval and the current solution.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval for the solve given by (t0, t1)
          dt (int, optional)
            The timestep for the solve. Defaults to length of interval

        *Returns*
          (timestep, solution_field) via (:py:class:`genexpr`)

        *Example of usage*::

          # Create generator
          solutions = solver.solve((0.0, 1.0), 0.1)

          # Iterate over generator (computes solutions as you go)
          for (interval, solution_fields) in solutions:
            (t0, t1) = interval
            v_, v = solution_fields
            # do something with the solutions
        """

        # Initial set-up
        # Solve on entire interval if no interval is given.
        (T0, T) = interval
        if dt is None:
            dt = T - T0
        t0 = T0
        t1 = T0 + dt

        # Step through time steps until at end time
        while True:
            info("Solving on t = (%g, %g)" % (t0, t1))
            self.step((t0, t1))

            # Yield solutions
            yield (t0, t1), self.solution_fields()

            # Break if this is the last step
            if end_of_time(T, t0, t1, dt):
                break

            # If not: update members and move to next time
            if isinstance(self.v_, backend.Function):
                self.v_.assign(self.v)
            else:
                debug("Assuming that v_ is updated elsewhere. Experimental.")

            t0 = t1
            t1 = t0 + dt

    def step(self, interval):
        """
        Solve on the given time interval (t0, t1).

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval (t0, t1) for the step

        *Invariants*
          Assuming that v\_ is in the correct state for t0, gives
          self.v in correct state at t1.
        """

        # Extract interval and thus time-step
        (t0, t1) = interval
        k_n = backend.Constant(t1 - t0)
        theta = self.parameters["theta"]

        # Extract conductivities
        M_i = self._M_i

        # Set time
        t = t0 + theta * (t1 - t0)
        self.time.assign(t)

        # Define variational formulation
        v = dolfin.TrialFunction(self.V)
        w = dolfin.TestFunction(self.V)
        Dt_v = (v - self.v_) / k_n
        v_mid = theta * v + (1.0 - theta) * self.v_

        (dz, rhs) = rhs_with_markerwise_field(self._I_s, self._mesh, w)
        theta_parabolic = ufl.inner(M_i * ufl.grad(v_mid), ufl.grad(w)) * dz()
        G = Dt_v * w * dz() + theta_parabolic - rhs

        # Define variational problem
        a, L = dolfin.system(G)
        pde = backend.LinearVariationalProblem(a, L, self.v)

        # Set-up solver
        solver = backend.LinearVariationalSolver(pde)
        solver.parameters.update(self.parameters["linear_variational_solver"])
        solver.solve()

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)

        To inspect all the default parameters, do::

          info(BasicMonodomainSolver.default_parameters(), True)
        """

        params = dolfin.Parameters("BasicMonodomainSolver")
        params.add("theta", 0.5)
        params.add("polynomial_degree", 1)
        params.add("family", "CG")
        params.add("enable_adjoint", True)

        params.add(backend.LinearVariationalSolver.default_parameters())
        return params


class MonodomainSolver(BasicMonodomainSolver):
    __doc__ = BasicMonodomainSolver.__doc__

    def __init__(self, mesh, time, M_i, I_s=None, v_=None, params=None):
        # Call super-class
        BasicMonodomainSolver.__init__(
            self, mesh, time, M_i, I_s=I_s, v_=v_, params=params
        )

        # Create variational forms
        self._timestep = backend.Constant(self.parameters["default_timestep"])
        (self._lhs, self._rhs, self._prec) = self.variational_forms(self._timestep)

        # Preassemble left-hand side (will be updated if time-step
        # changes)
        debug("Preassembling monodomain matrix (and initializing vector)")
        self._lhs_matrix = backend.assemble(self._lhs, **self._annotate_kwargs)
        self._rhs_vector = dolfin.Vector(mesh.mpi_comm(), self._lhs_matrix.size(0))
        self._lhs_matrix.init_vector(self._rhs_vector, 0)

        # Create linear solver (based on parameter choices)
        self._linear_solver, self._update_solver = self._create_linear_solver()

    @property
    def linear_solver(self):
        """The linear solver (:py:class:`dolfin.LUSolver` or
        :py:class:`dolfin.KrylovSolver`)."""
        return self._linear_solver

    def _create_linear_solver(self):
        "Helper function for creating linear solver based on parameters."
        solver_type = self.parameters["linear_solver_type"]

        if solver_type == "direct":
            solver = backend.LUSolver(self._lhs_matrix, self.parameters["lu_type"])
            solver.parameters.update(self.parameters["lu_solver"])
            update_routine = self._update_lu_solver

        elif solver_type == "iterative":
            # Preassemble preconditioner (will be updated if time-step
            # changes)
            debug("Preassembling preconditioner")
            # Initialize KrylovSolver with matrix and preconditioner
            alg = self.parameters["algorithm"]
            prec = self.parameters["preconditioner"]
            if self.parameters["use_custom_preconditioner"]:
                self._prec_matrix = backend.assemble(
                    self._prec, **self._annotate_kwargs
                )
                solver = backend.PETScKrylovSolver(alg, prec)
                solver.parameters.update(self.parameters["krylov_solver"])
                solver.set_operators(self._lhs_matrix, self._prec_matrix)
                solver.ksp().setFromOptions()
            else:
                solver = backend.PETScKrylovSolver(alg, prec)
                solver.parameters.update(self.parameters["krylov_solver"])
                solver.set_operator(self._lhs_matrix)
                solver.ksp().setFromOptions()

            update_routine = self._update_krylov_solver
        else:
            error("Unknown linear_solver_type given: %s" % solver_type)

        return (solver, update_routine)

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)

        To inspect all the default parameters, do::

          info(MonodomainSolver.default_parameters(), True)
        """

        params = dolfin.Parameters("MonodomainSolver")
        params.add("enable_adjoint", True)
        params.add("theta", 0.5)
        params.add("polynomial_degree", 1)
        params.add("default_timestep", 1.0)
        params.add("family", "CG")

        # Set default solver type to be iterative
        params.add("linear_solver_type", "iterative")
        params.add("lu_type", "default")

        # Set default iterative solver choices (used if iterative
        # solver is invoked)
        params.add("algorithm", "cg")
        params.add("preconditioner", "petsc_amg")
        params.add("use_custom_preconditioner", True)

        # Add default parameters from both LU and Krylov solvers
        params.add(backend.LUSolver.default_parameters())
        params.add(backend.KrylovSolver.default_parameters())

        # Customize default parameters for KrylovSolver
        # params["krylov_solver"]["preconditioner"]["structure"] = "same"

        return params

    def variational_forms(self, k_n):
        """Create the variational forms corresponding to the given
        discretization of the given system of equations.

        *Arguments*
          k_n (:py:class:`ufl.Expr` or float)
            The time step

        *Returns*
          (lhs, rhs, prec) (:py:class:`tuple` of :py:class:`ufl.Form`)

        """

        # Extract theta parameter and conductivities
        theta = self.parameters["theta"]
        M_i = self._M_i

        # Define variational formulation
        v = dolfin.TrialFunction(self.V)
        w = dolfin.TestFunction(self.V)

        # Set-up variational problem
        Dt_v_k_n = v - self.v_
        v_mid = theta * v + (1.0 - theta) * self.v_

        (dz, rhs) = rhs_with_markerwise_field(self._I_s, self._mesh, w)
        theta_parabolic = ufl.inner(M_i * ufl.grad(v_mid), ufl.grad(w)) * dz()
        G = Dt_v_k_n * w * dz + k_n * theta_parabolic - k_n * rhs

        # Define preconditioner based on educated(?) guess by Marie
        prec = (v * w + k_n / 2.0 * ufl.inner(M_i * ufl.grad(v), ufl.grad(w))) * dz

        (a, L) = dolfin.system(G)
        return (a, L, prec)

    def step(self, interval):
        """
        Solve on the given time step (t0, t1).

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval (t0, t1) for the step

        *Invariants*
          Assuming that v\_ is in the correct state for t0, gives
          self.v in correct state at t1.
        """

        timer = dolfin.Timer("PDE Step")

        # Extract interval and thus time-step
        (t0, t1) = interval
        dt = t1 - t0
        theta = self.parameters["theta"]
        t = t0 + theta * dt
        self.time.assign(t)

        # Update matrix and linear solvers etc as needed
        timestep_unchanged = abs(dt - float(self._timestep)) < 1.0e-12
        self._update_solver(timestep_unchanged, dt)

        # Assemble right-hand-side
        timer0 = dolfin.Timer("Assemble rhs")
        backend.assemble(self._rhs, tensor=self._rhs_vector, **self._annotate_kwargs)
        del timer0

        # Solve problem
        self.linear_solver.solve(
            self.v.vector(), self._rhs_vector, **self._annotate_kwargs
        )
        timer.stop()

    def _update_lu_solver(self, timestep_unchanged, dt):
        """Helper function for updating an LUSolver depending on
        whether timestep has changed."""

        # Update stored timestep
        # FIXME: dolfin_adjoint still can't annotate constant assignment.
        self._timestep.assign(backend.Constant(dt))  # , annotate=annotate)

        # Reassemble matrix
        backend.assemble(self._lhs, tensor=self._lhs_matrix, **self._annotate_kwargs)

    def _update_krylov_solver(self, timestep_unchanged, dt):
        """Helper function for updating a KrylovSolver depending on
        whether timestep has changed."""

        annotate_kwargs(self.parameters)
        # Update reuse of preconditioner parameter in accordance with
        # changes in timestep
        if timestep_unchanged:
            debug("Timestep is unchanged, reusing preconditioner")
            # self.linear_solver.parameters["preconditioner"]["structure"] = "same"
        else:
            debug("Timestep has changed, updating preconditioner")
            # self.linear_solver.parameters["preconditioner"]["structure"] = \
            #                                            "same_nonzero_pattern"

            # Update stored timestep
            self._timestep.assign(backend.Constant(dt))

            # Reassemble matrix
            backend.assemble(
                self._lhs, tensor=self._lhs_matrix, **self._annotate_kwargs
            )

            # Reassemble preconditioner
            if self.parameters["use_custom_preconditioner"]:
                backend.assemble(
                    self._prec, tensor=self._prec_matrix, **self._annotate_kwargs
                )

        # Set nonzero initial guess if it indeed is nonzero
        # if (self.v.vector().norm("l2") > 1.e-12):
        #    debug("Initial guess is non-zero.")
        #    self.linear_solver.parameters["nonzero_initial_guess"] = True



"This module contains solvers for (subclasses of) CardiacCellModel."

__author__ = "Marie E. Rognes (meg@simula.no), 2012--2013"

__all__ = [
    "BasicSingleCellSolver",
    "BasicCardiacODESolver",
    "CardiacODESolver",
    "SingleCellSolver",
]

import dolfin
from dolfin import (
    ForwardEuler,  # noqa:F401
    BackwardEuler,  # noqa:F401
    CrankNicolson,  # noqa:F401
    RK4,  # noqa:F401
    ESDIRK3,  # noqa:F401
    ESDIRK4,  # noqa:F401
    RL1,  # noqa:F401
    RL2,  # noqa:F401
    GRL1,  # noqa:F401
    GRL2,  # noqa:F401
)
from cbcbeat.dolfinimport import backend
from cbcbeat import CardiacCellModel, MultiCellModel
from cbcbeat.markerwisefield import (
    handle_markerwise,
    Markerwise,
    rhs_with_markerwise_field,
)
from cbcbeat.utils import state_space, TimeStepper, splat, annotate_kwargs
import ufl
import ufl.classes
import cbcbeat
from ufl.log import info_blue, error


def point_integral_solver_default_parameters():
    try:
        p = backend.PointIntegralSolver.default_parameters()
    except AttributeError:
        p = dolfin.Parameters("point_integral_solver")
        p.add("reset_stage_solutions", True)
        # Set parameters for NewtonSolver
        pn = dolfin.Parameters("newton_solver")
        pn.add("maximum_iterations", 40)
        pn.add("always_recompute_jacobian", False)
        pn.add("recompute_jacobian_each_solve", True)
        pn.add("relaxation_parameter", 1.0, 0.0, 1.0)
        pn.add("relative_tolerance", 1e-10, 1e-20, 2.0)
        pn.add("absolute_tolerance", 1e-15, 1e-20, 2.0)
        pn.add("kappa", 0.1, 0.05, 1.0)
        pn.add("eta_0", 1.0, 1e-15, 1.0)
        pn.add("max_relative_previous_residual", 1e-1, 1e-5, 1.0)
        pn.add("reset_each_step", True)
        pn.add("report", False)
        pn.add("report_vertex", 0, 0, 32767)
        pn.add("verbose_report", False)
        p.add(pn)
    return p


class BasicCardiacODESolver(object):
    """A basic, non-optimised solver for systems of ODEs typically
    encountered in cardiac applications of the form: find a scalar
    field :math:`v = v(x, t)` and a vector field :math:`s = s(x, t)`

    .. math::

      v_t = - I_{ion}(v, s) + I_s

      s_t = F(v, s)

    where :math:`I_{ion}` and :math:`F` are given non-linear
    functions, and :math:`I_s` is some prescribed stimulus.

    Here, this nonlinear ODE system is solved via a theta-scheme.  By
    default theta=0.5, which corresponds to a Crank-Nicolson
    scheme. This can be changed by modifying the solver parameters.

    .. note::

       For the sake of simplicity and consistency with other solver
       objects, this solver operates on its solution fields (as state
       variables) directly internally. More precisely, solve (and
       step) calls will act by updating the internal solution
       fields. It implies that initial conditions can be set (and are
       intended to be set) by modifying the solution fields prior to
       simulation.

    *Arguments*
      mesh (:py:class:`dolfin.Mesh`)
        The spatial domain (mesh)

      time (:py:class:`dolfin.Constant` or None)
        A constant holding the current time. If None is given, time is
        created for you, initialized to zero.

      model (:py:class:`cbcbeat.CardiacCellModel`)
        A representation of the cardiac cell model(s)

      I_s (optional) A typically time-dependent external stimulus
        given as a :py:class:`dolfin.cpp.function.GenericFunction` or a
        Markerwise. NB: it is assumed that the time dependence of I_s
        is encoded via the 'time' Constant.

      params (:py:class:`dolfin.Parameters`, optional)
        Solver parameters
    """

    def __init__(self, mesh, time, model, I_s=None, params=None):
        # Store input
        self._mesh = mesh
        self._time = time
        self._model = model

        # Extract some information from cell model
        self._F = self._model.F
        self._I_ion = self._model.I
        self._num_states = self._model.num_states()

        # Handle stimulus
        self._I_s = handle_markerwise(I_s, dolfin.cpp.function.GenericFunction)

        # Initialize and update parameters if given
        self.parameters = self.default_parameters()
        if params is not None:
            self.parameters.update(params)

        # Create (mixed) function space for potential + states
        v_family = self.parameters["V_polynomial_family"]
        v_degree = self.parameters["V_polynomial_degree"]
        s_family = self.parameters["S_polynomial_family"]
        s_degree = self.parameters["S_polynomial_degree"]

        if v_family == s_family and s_degree == v_degree:
            self.VS = dolfin.VectorFunctionSpace(
                self._mesh, v_family, v_degree, dim=self._num_states + 1
            )
        else:
            V = dolfin.FunctionSpace(self._mesh, v_family, v_degree)
            S = state_space(self._mesh, self._num_states, s_family, s_degree)
            Mx = dolfin.MixedElement(V.ufl_element(), S.ufl_element())
            self.VS = dolfin.FunctionSpace(self._mesh, Mx)

        # Initialize solution fields
        self.vs_ = backend.Function(self.VS, name="vs_")
        self.vs = backend.Function(self.VS, name="vs")

    @property
    def time(self):
        "The internal time of the solver."
        return self._time

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)
        """
        params = dolfin.Parameters("BasicCardiacODESolver")
        params.add("theta", 0.5)
        params.add("V_polynomial_degree", 0)
        params.add("V_polynomial_family", "DG")
        params.add("S_polynomial_degree", 0)
        params.add("S_polynomial_family", "DG")
        params.add("enable_adjoint", True)

        # Use iterative solver as default.
        params.add(backend.NonlinearVariationalSolver.default_parameters())
        params["nonlinear_variational_solver"]["newton_solver"][
            "linear_solver"
        ] = "gmres"

        return params

    def solution_fields(self):
        """
        Return tuple of previous and current solution objects.

        Modifying these will modify the solution objects of the solver
        and thus provides a way for setting initial conditions for
        instance.

        *Returns*
          (previous vs, current vs) (:py:class:`tuple` of :py:class:`dolfin.Function`)
        """
        return (self.vs_, self.vs)

    def solve(self, interval, dt=None):
        """
        Solve the problem given by the model on a given time interval
        (t0, t1) with a given timestep dt and return generator for a
        tuple of the interval and the current vs solution.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval for the solve given by (t0, t1)
          dt (int, optional)
            The timestep for the solve. Defaults to length of interval

        *Returns*
          (timestep, current vs) via (:py:class:`genexpr`)

        *Example of usage*::

          # Create generator
          solutions = solver.solve((0.0, 1.0), 0.1)

          # Iterate over generator (computes solutions as you go)
          for (interval, vs) in solutions:
            # do something with the solutions

        """

        # Initial time set-up
        (T0, T) = interval

        # Solve on entire interval if no interval is given.
        if dt is None:
            dt = T - T0

        # Create timestepper
        time_stepper = TimeStepper(
            interval, dt, annotate=self.parameters["enable_adjoint"]
        )
        for t0, t1 in time_stepper:
            info_blue("Solving on t = (%g, %g)" % (t0, t1))
            self.step((t0, t1))

            # Yield solutions
            yield (t0, t1), self.vs

            self.vs_.assign(self.vs)

    def step(self, interval):
        """
        Solve on the given time step (t0, t1).

        End users are recommended to use solve instead.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval (t0, t1) for the step
        """

        timer = dolfin.Timer("ODE step")

        # Check for cell meshs
        self._mesh.topology().dim()

        # Extract time mesh
        (t0, t1) = interval
        k_n = backend.Constant(t1 - t0)

        # Extract previous solution(s)
        (v_, s_) = splat(self.vs_, self._num_states + 1)

        # Set-up current variables
        self.vs.assign(self.vs_)  # Start with good guess
        (v, s) = splat(self.vs, self._num_states + 1)
        (w, r) = splat(dolfin.TestFunction(self.VS), self._num_states + 1)

        # Define equation based on cell model
        Dt_v = (v - v_) / k_n
        Dt_s = (s - s_) / k_n

        theta = self.parameters["theta"]

        # Set time (propagates to time-dependent variables defined via
        # self.time)
        t = t0 + theta * (t1 - t0)
        self.time.assign(t)

        v_mid = theta * v + (1.0 - theta) * v_
        s_mid = theta * s + (1.0 - theta) * s_

        if isinstance(self._model, MultiCellModel):
            # assert(model.mesh() == self._mesh)

            model = self._model
            mesh = model.mesh()
            dy = dolfin.Measure("dx", domain=mesh, subdomain_data=model.markers())

            # Only allowing trivial forcing functions here
            if isinstance(self._I_s, Markerwise):
                error("Not implemented")
            rhs = self._I_s * w * dy()

            n = model.num_states()  # Extract number of global states

            # Collect contributions to lhs by iterating over the different cell models
            domains = self._model.keys()
            lhs = list()
            for k, model_k in enumerate(model.models()):
                n_k = (
                    model_k.num_states()
                )  # Extract number of local (non-trivial) states

                # Extract right components of coefficients and test functions
                # () is not the same as (1,)
                if n_k == 1:
                    s_mid_k = s_mid[0]
                    r_k = r[0]
                    Dt_s_k = Dt_s[0]
                else:
                    s_mid_k = dolfin.as_vector(tuple(s_mid[j] for j in range(n_k)))
                    r_k = dolfin.as_vector(tuple(r[j] for j in range(n_k)))
                    Dt_s_k = dolfin.as_vector(tuple(Dt_s[j] for j in range(n_k)))

                i_k = domains[k]  # Extract domain index of cell model k

                # Extract right currents and ion channel expressions
                F_theta_k = self._F(v_mid, s_mid_k, time=self.time, index=i_k)
                I_theta_k = -self._I_ion(v_mid, s_mid_k, time=self.time, index=i_k)

                # Variational contribution over the relevant domain
                a_k = (
                    (Dt_v - I_theta_k) * w
                    + dolfin.inner(Dt_s_k, r_k)
                    + dolfin.inner(-F_theta_k, r_k)
                ) * dy(i_k)

                # Add s_trivial = 0 on Omega_{i_k} in variational form:
                a_k += sum(s[j] * r[j] for j in range(n_k, n)) * dy(i_k)
                lhs.append(a_k)
            lhs = sum(lhs)

        else:
            (dz, rhs) = rhs_with_markerwise_field(self._I_s, self._mesh, w)

            # Evaluate currents at averaged v and s. Note sign for I_theta
            F_theta = self._F(v_mid, s_mid, time=self.time)
            I_theta = -self._I_ion(v_mid, s_mid, time=self.time)
            lhs = (Dt_v - I_theta) * w * dz + dolfin.inner(Dt_s - F_theta, r) * dz

        # Set-up system of equations
        G = lhs - rhs

        # Solve system
        pde = backend.NonlinearVariationalProblem(
            G, self.vs, J=dolfin.derivative(G, self.vs)
        )
        solver = backend.NonlinearVariationalSolver(pde)
        solver_params = self.parameters["nonlinear_variational_solver"]
        solver.parameters.update(solver_params)
        solver.solve()
        timer.stop()


class CardiacODESolver(object):
    """An optimised solver for systems of ODEs typically
    encountered in cardiac applications of the form: find a scalar
    field :math:`v = v(x, t)` and a vector field :math:`s = s(x, t)`

    .. math::

      v_t = - I_{ion}(v, s) + I_s

      s_t = F(v, s)

    where :math:`I_{ion}` and :math:`F` are given non-linear
    functions, and :math:`I_s` is some prescribed stimulus.

    .. note::

       For the sake of simplicity and consistency with other solver
       objects, this solver operates on its solution fields (as state
       variables) directly internally. More precisely, solve (and
       step) calls will act by updating the internal solution
       fields. It implies that initial conditions can be set (and are
       intended to be set) by modifying the solution fields prior to
       simulation.

    *Arguments*
      mesh (:py:class:`dolfin.Mesh`)
        The spatial mesh (mesh)

      time (:py:class:`dolfin.Constant` or None)
        A constant holding the current time. If None is given, time is
        created for you, initialized to zero.

      model (:py:class:`cbcbeat.CardiacCellModel`)
        A representation of the cardiac cell model(s)

      I_s (:py:class:`dolfin.Expression`, optional)
        A typically time-dependent external stimulus. NB: it is
        assumed that the time dependence of I_s is encoded via the
        'time' Constant.

      params (:py:class:`dolfin.Parameters`, optional)
        Solver parameters

    """

    def __init__(self, mesh, time, model, I_s=None, params=None):
        # Store input
        self._mesh = mesh
        self._time = time
        self._model = model

        # Extract some information from cell model
        self._F = self._model.F
        self._I_ion = self._model.I
        self._num_states = self._model.num_states()

        self._I_s = handle_markerwise(I_s, dolfin.cpp.function.GenericFunction)

        # Create time if not given, otherwise use given time
        if time is None:
            self._time = backend.Constant(0.0)
        else:
            self._time = time

        # Initialize and update parameters if given
        self.parameters = self.default_parameters()
        if params is not None:
            self.parameters.update(params)

        # Create (vector) function space for potential + states
        self.VS = dolfin.VectorFunctionSpace(
            self._mesh,
            "CG",
            self.parameters["polynomial_degree"],
            dim=self._num_states + 1,
        )

        # Initialize solution field
        self.vs_ = backend.Function(self.VS, name="vs_")
        self.vs = backend.Function(self.VS, name="vs")

        # Initialize scheme
        (v, s) = splat(self.vs, self._num_states + 1)
        (w, q) = splat(dolfin.TestFunction(self.VS), self._num_states + 1)

        # Workaround to get algorithm in RL schemes working as it only
        # works for scalar expressions
        F_exprs = self._F(v, s, self._time)

        # MER: This looks much more complicated than it needs to be!
        # If we have a dolfin.as_vector expression
        F_exprs_q = ufl.zero()
        if isinstance(F_exprs, ufl.classes.ListTensor):
            # for i, expr_i in enumerate(F_exprs.operands()):
            for i, expr_i in enumerate(F_exprs.ufl_operands):
                F_exprs_q += expr_i * q[i]
        else:
            F_exprs_q = F_exprs * q

        rhs = F_exprs_q - self._I_ion(v, s, self._time) * w

        # Handle stimulus: only handle single function case for now
        msg = "Markerwise stimulus not supported by PointIntegralSolver."
        assert not isinstance(self._I_s, Markerwise), msg
        if self._I_s:
            rhs += self._I_s * w

        # FIXME: The application of dP was moved so adding an integral
        # is done just once. Otherwise ufl could not figure out that
        # we had only one integral...
        self._rhs = rhs * dolfin.dP()

        # sys.exit()
        name = self.parameters["scheme"]
        Scheme = self._name_to_scheme(name)
        self._scheme = Scheme(self._rhs, self.vs, self._time)

        # Figure out whether we should annotate or not
        self._annotate_kwargs = annotate_kwargs(self.parameters)

        # Initialize solver and update its parameters
        self._pi_solver = backend.PointIntegralSolver(self._scheme)
        self._pi_solver.parameters.update(self.parameters["point_integral_solver"])

    def _name_to_scheme(self, name):
        """Return scheme class with given name

        *Arguments*
          name (string)

        *Returns*
          the Scheme (:py:class:`dolfin.MultiStageScheme`)

        """
        return eval(name)

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)
        """
        params = dolfin.Parameters("CardiacODESolver")
        params.add("scheme", "BackwardEuler")
        params.add("polynomial_degree", 1)
        params.add(point_integral_solver_default_parameters())
        params.add("enable_adjoint", True)

        return params

    def solution_fields(self):
        """
        Return current solution object.

        Modifying this will modify the solution object of the solver
        and thus provides a way for setting initial conditions for
        instance.

        *Returns*
          (previous vs_, current vs) (:py:class:`dolfin.Function`)
        """
        return (self.vs_, self.vs)

    def step(self, interval):
        """
        Solve on the given time step (t0, t1).

        End users are recommended to use solve instead.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval (t0, t1) for the step
        """
        # NB: The point integral solver operates on vs directly, map
        # initial condition in vs_ to vs:

        timer = dolfin.Timer("ODE step")
        self.vs.assign(self.vs_)

        (t0, t1) = interval
        dt = t1 - t0

        self._annotate_kwargs = annotate_kwargs(self.parameters)
        if cbcbeat.dolfinimport.has_dolfin_adjoint:
            self._pi_solver.step(dt, **self._annotate_kwargs)
        else:
            self._pi_solver.step(dt)
        timer.stop()

    def solve(self, interval, dt=None):
        """
        Solve the problem given by the model on a given time interval
        (t0, t1) with a given timestep dt and return generator for a
        tuple of the interval and the current vs solution.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval for the solve given by (t0, t1)
          dt (int, optional)
            The timestep for the solve. Defaults to length of interval

        *Returns*
          (timestep, current vs) via (:py:class:`genexpr`)

        *Example of usage*::

          # Create generator
          solutions = solver.solve((0.0, 1.0), 0.1)

          # Iterate over generator (computes solutions as you go)
          for (interval, vs) in solutions:
            # do something with the solutions

        """

        # Initial time set-up
        (T0, T) = interval

        # Solve on entire interval if no interval is given.
        if dt is None:
            dt = T - T0

        # Create timestepper
        time_stepper = TimeStepper(
            interval, dt, annotate=self.parameters["enable_adjoint"]
        )

        for t0, t1 in time_stepper:
            info_blue("Solving on t = (%g, %g)" % (t0, t1))
            self.step((t0, t1))

            # Yield solutions
            yield (t0, t1), self.vs

            # FIXME: This eventually breaks in parallel!?
            self.vs_.assign(self.vs)


class BasicSingleCellSolver(BasicCardiacODESolver):
    """A basic, non-optimised solver for systems of ODEs typically
    encountered in cardiac applications of the form: find a scalar
    field :math:`v = v(t)` and a vector field :math:`s = s(t)`

    .. math::

      v_t = - I_{ion}(v, s) + I_s

      s_t = F(v, s)

    where :math:`I_{ion}` and :math:`F` are given non-linear
    functions, :math:`I_s` is some prescribed stimulus. If :math:`I_s`
    depends on time, it is assumed that :math:`I_s` is a
    :py:class:`dolfin.Expression` with parameter 't'.

    Use this solver if you just want to test the results from a
    cardiac cell model without any spatial mesh dependence.

    Here, this nonlinear ODE system is solved via a theta-scheme.  By
    default theta=0.5, which corresponds to a Crank-Nicolson
    scheme. This can be changed by modifying the solver parameters.

    .. note::

       For the sake of simplicity and consistency with other solver
       objects, this solver operates on its solution fields (as state
       variables) directly internally. More precisely, solve (and
       step) calls will act by updating the internal solution
       fields. It implies that initial conditions can be set (and are
       intended to be set) by modifying the solution fields prior to
       simulation.

    *Arguments*
      model (:py:class:`~cbcbeat.cellmodels.cardiaccellmodel.CardiacCellModel`)
        A cardiac cell model
      time (:py:class:`~dolfin.Constant` or None)
        A constant holding the current time.
      params (:py:class:`dolfin.Parameters`, optional)
        Solver parameters

    """

    def __init__(self, model, time, params=None):
        "Create solver from given cell model and optional parameters."

        assert isinstance(model, CardiacCellModel), (
            "Expecting model to be a CardiacCellModel, not %r" % model
        )
        assert isinstance(time, backend.Constant), (
            "Expecting time to be a Constant instance, not %r" % time
        )
        assert isinstance(params, dolfin.Parameters) or params is None, (
            "Expecting params to be a Parameters (or None), not %r" % params
        )

        # Store model
        self._model = model

        # Define carefully chosen dummy mesh
        mesh = dolfin.UnitIntervalMesh(1)

        # Extract information from cardiac cell model and ship off to
        # super-class.
        BasicCardiacODESolver.__init__(
            self, mesh, time, model, I_s=model.stimulus, params=params
        )


class SingleCellSolver(CardiacODESolver):
    def __init__(self, model, time, params=None):
        "Create solver from given cell model and optional parameters."

        assert isinstance(model, CardiacCellModel), (
            "Expecting model to be a CardiacCellModel, not %r" % model
        )
        assert isinstance(time, backend.Constant), (
            "Expecting time to be a Constant instance, not %r" % time
        )
        assert isinstance(params, dolfin.Parameters) or params is None, (
            "Expecting params to be a Parameters (or None), not %r" % params
        )

        # Store model
        self._model = model

        # Define carefully chosen dummy mesh
        mesh = dolfin.UnitIntervalMesh(1)

        # Extract information from cardiac cell model and ship off to
        # super-class.
        CardiacODESolver.__init__(
            self, mesh, time, model, I_s=model.stimulus, params=params
        )



"""This module contains a dummy cardiac cell model."""

__author__ = "Marie E. Rognes (meg@simula.no), 2012--2013"
__all__ = ["NoCellModel"]

from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel
from collections import OrderedDict

# FIXME: This class represents a design flaw rather than anything
# else. Remove in a clean-up of the solvers.


class NoCellModel(CardiacCellModel):
    """
    Class representing no cell model (only bidomain equations). It
    actually just represents a single completely decoupled ODE.
    """

    def __init__(self, params=None, init_conditions=None):
        CardiacCellModel.__init__(self, params, init_conditions)

    def I(self, v, s, time=None):
        return 0

    def F(self, v, s, time=None):
        return -s

    def num_states(self):
        return 1

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict([("V", 0.0), ("S", 0.0)])
        return ic

    def __str__(self):
        return "No cardiac cell model"



"""This module contains a Beeler_reuter_1977 cardiac cell model

The module was autogenerated from a gotran ode file
"""
from __future__ import division
from collections import OrderedDict
import ufl

import dolfin
from cbcbeat.dolfinimport import backend
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class Beeler_reuter_1977(CardiacCellModel):
    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        """
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("E_Na", 50),
                ("g_Na", 0.04),
                ("g_Nac", 3e-05),
                ("g_s", 0.0009),
                ("IstimAmplitude", 0.0),
                ("IstimEnd", 50000),
                ("IstimPeriod", 1000),
                ("IstimPulseDuration", 1),
                ("IstimStart", 1),
                ("C", 0.01),
            ]
        )
        return params

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict(
            [
                ("V", -84.624),
                ("m", 0.011),
                ("h", 0.988),
                ("j", 0.975),
                ("Cai", 0.0001),
                ("d", 0.003),
                ("f", 0.994),
                ("x1", 0.0001),
            ]
        )
        return ic

    def _I(self, v, s, time):
        """
        Original gotran transmembrane current dV/dt
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 7
        m, h, j, Cai, d, f, x1 = s

        # Assign parameters
        E_Na = self._parameters["E_Na"]
        g_Na = self._parameters["g_Na"]
        g_Nac = self._parameters["g_Nac"]
        g_s = self._parameters["g_s"]
        IstimAmplitude = self._parameters["IstimAmplitude"]
        IstimPulseDuration = self._parameters["IstimPulseDuration"]
        IstimStart = self._parameters["IstimStart"]
        C = self._parameters["C"]

        # Init return args
        current = [ufl.zero()] * 1

        # Expressions for the Sodium current component
        i_Na = (g_Nac + g_Na * (m * m * m) * h * j) * (-E_Na + V)

        # Expressions for the Slow inward current component
        E_s = -82.3 - 13.0287 * ufl.ln(0.001 * Cai)
        i_s = g_s * (-E_s + V) * d * f

        # Expressions for the Time dependent outward current component
        i_x1 = (
            0.00197277571153
            * (-1 + 21.7584023962 * ufl.exp(0.04 * V))
            * ufl.exp(-0.04 * V)
            * x1
        )

        # Expressions for the Time independent outward current component
        i_K1 = 0.0035 * (-4 + 119.85640019 * ufl.exp(0.04 * V)) / (
            8.33113748769 * ufl.exp(0.04 * V) + 69.4078518388 * ufl.exp(0.08 * V)
        ) + 0.0035 * (4.6 + 0.2 * V) / (1 - 0.398519041085 * ufl.exp(-0.04 * V))

        # Expressions for the Stimulus protocol component
        Istim = ufl.conditional(
            ufl.And(
                ufl.ge(time, IstimStart), ufl.le(time, IstimPulseDuration + IstimStart)
            ),
            IstimAmplitude,
            0,
        )

        # Expressions for the Membrane component
        current[0] = (-i_K1 + Istim - i_Na - i_x1 - i_s) / C

        # Return results
        return current[0]

    def I(self, v, s, time=None):
        """
        Transmembrane current

           I = -dV/dt

        """
        return -self._I(v, s, time)

    def F(self, v, s, time=None):
        """
        Right hand side for ODE system
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 7
        m, h, j, Cai, d, f, x1 = s

        # Assign parameters
        g_s = self._parameters["g_s"]

        # Init return args
        F_expressions = [ufl.zero()] * 7

        # Expressions for the Sodium current m gate component
        alpha_m = (-47 - V) / (-1 + 0.0090952771017 * ufl.exp(-0.1 * V))
        beta_m = 0.709552672749 * ufl.exp(-0.056 * V)
        F_expressions[0] = -beta_m * m + (1 - m) * alpha_m

        # Expressions for the Sodium current h gate component
        alpha_h = 5.49796243871e-10 * ufl.exp(-0.25 * V)
        beta_h = 1.7 / (1 + 0.15802532089 * ufl.exp(-0.082 * V))
        F_expressions[1] = (1 - h) * alpha_h - beta_h * h

        # Expressions for the Sodium current j gate component
        alpha_j = (
            1.86904730072e-10
            * ufl.exp(-0.25 * V)
            / (1 + 1.67882753e-07 * ufl.exp(-0.2 * V))
        )
        beta_j = 0.3 / (1 + 0.0407622039784 * ufl.exp(-0.1 * V))
        F_expressions[2] = (1 - j) * alpha_j - beta_j * j

        # Expressions for the Slow inward current component
        E_s = -82.3 - 13.0287 * ufl.ln(0.001 * Cai)
        i_s = g_s * (-E_s + V) * d * f
        F_expressions[3] = 7e-06 - 0.07 * Cai - 0.01 * i_s

        # Expressions for the Slow inward current d gate component
        alpha_d = (
            0.095
            * ufl.exp(1 / 20 - V / 100)
            / (1 + 1.43328813857 * ufl.exp(-0.0719942404608 * V))
        )
        beta_d = 0.07 * ufl.exp(-44 / 59 - V / 59) / (1 + ufl.exp(11 / 5 + V / 20))
        F_expressions[4] = -beta_d * d + (1 - d) * alpha_d

        # Expressions for the Slow inward current f gate component
        alpha_f = (
            0.012
            * ufl.exp(-28 / 125 - V / 125)
            / (1 + 66.5465065251 * ufl.exp(0.149925037481 * V))
        )
        beta_f = 0.0065 * ufl.exp(-3 / 5 - V / 50) / (1 + ufl.exp(-6 - V / 5))
        F_expressions[5] = (1 - f) * alpha_f - beta_f * f

        # Expressions for the Time dependent outward current x1 gate component
        alpha_x1 = (
            0.0311584109863
            * ufl.exp(0.0826446280992 * V)
            / (1 + 17.4117080633 * ufl.exp(0.0571428571429 * V))
        )
        beta_x1 = (
            0.000391646440562
            * ufl.exp(-0.0599880023995 * V)
            / (1 + ufl.exp(-4 / 5 - V / 25))
        )
        F_expressions[6] = (1 - x1) * alpha_x1 - beta_x1 * x1

        # Return results
        return dolfin.as_vector(F_expressions)

    def num_states(self):
        return 7

    def __str__(self):
        return "Beeler_reuter_1977 cardiac cell model"



"""This module contains a Tentusscher_panfilov_2006_epi_cell cardiac cell model

The module was autogenerated from a gotran ode file
"""
from __future__ import division
from collections import OrderedDict
import ufl
import dolfin

from cbcbeat.dolfinimport import backend
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class Tentusscher_panfilov_2006_epi_cell(CardiacCellModel):
    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        """
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("P_kna", 0.03),
                ("g_K1", 5.405),
                ("g_Kr", 0.153),
                ("g_Ks", 0.392),
                ("g_Na", 14.838),
                ("g_bna", 0.00029),
                ("g_CaL", 3.98e-05),
                ("g_bca", 0.000592),
                ("g_to", 0.294),
                ("K_mNa", 40),
                ("K_mk", 1),
                ("P_NaK", 2.724),
                ("K_NaCa", 1000),
                ("K_sat", 0.1),
                ("Km_Ca", 1.38),
                ("Km_Nai", 87.5),
                ("alpha", 2.5),
                ("gamma", 0.35),
                ("K_pCa", 0.0005),
                ("g_pCa", 0.1238),
                ("g_pK", 0.0146),
                ("Buf_c", 0.2),
                ("Buf_sr", 10),
                ("Buf_ss", 0.4),
                ("Ca_o", 2),
                ("EC", 1.5),
                ("K_buf_c", 0.001),
                ("K_buf_sr", 0.3),
                ("K_buf_ss", 0.00025),
                ("K_up", 0.00025),
                ("V_leak", 0.00036),
                ("V_rel", 0.102),
                ("V_sr", 0.001094),
                ("V_ss", 5.468e-05),
                ("V_xfer", 0.0038),
                ("Vmax_up", 0.006375),
                ("k1_prime", 0.15),
                ("k2_prime", 0.045),
                ("k3", 0.06),
                ("k4", 0.005),
                ("max_sr", 2.5),
                ("min_sr", 1),
                ("Na_o", 140),
                ("Cm", 0.185),
                ("F", 96485.3415),
                ("R", 8314.472),
                ("T", 310),
                ("V_c", 0.016404),
                ("stim_amplitude", 0),
                ("stim_duration", 1),
                ("stim_period", 1000),
                ("stim_start", 10),
                ("K_o", 5.4),
            ]
        )
        return params

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict(
            [
                ("V", -85.23),
                ("Xr1", 0.00621),
                ("Xr2", 0.4712),
                ("Xs", 0.0095),
                ("m", 0.00172),
                ("h", 0.7444),
                ("j", 0.7045),
                ("d", 3.373e-05),
                ("f", 0.7888),
                ("f2", 0.9755),
                ("fCass", 0.9953),
                ("s", 0.999998),
                ("r", 2.42e-08),
                ("Ca_i", 0.000126),
                ("R_prime", 0.9073),
                ("Ca_SR", 3.64),
                ("Ca_ss", 0.00036),
                ("Na_i", 8.604),
                ("K_i", 136.89),
            ]
        )
        return ic

    def _I(self, v, s, time):
        """
        Original gotran transmembrane current dV/dt
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 18
        (
            Xr1,
            Xr2,
            Xs,
            m,
            h,
            j,
            d,
            f,
            f2,
            fCass,
            s,
            r,
            Ca_i,
            R_prime,
            Ca_SR,
            Ca_ss,
            Na_i,
            K_i,
        ) = s

        # Assign parameters
        P_kna = self._parameters["P_kna"]
        g_K1 = self._parameters["g_K1"]
        g_Kr = self._parameters["g_Kr"]
        g_Ks = self._parameters["g_Ks"]
        g_Na = self._parameters["g_Na"]
        g_bna = self._parameters["g_bna"]
        g_CaL = self._parameters["g_CaL"]
        g_bca = self._parameters["g_bca"]
        g_to = self._parameters["g_to"]
        K_mNa = self._parameters["K_mNa"]
        K_mk = self._parameters["K_mk"]
        P_NaK = self._parameters["P_NaK"]
        K_NaCa = self._parameters["K_NaCa"]
        K_sat = self._parameters["K_sat"]
        Km_Ca = self._parameters["Km_Ca"]
        Km_Nai = self._parameters["Km_Nai"]
        alpha = self._parameters["alpha"]
        gamma = self._parameters["gamma"]
        K_pCa = self._parameters["K_pCa"]
        g_pCa = self._parameters["g_pCa"]
        g_pK = self._parameters["g_pK"]
        Ca_o = self._parameters["Ca_o"]
        Na_o = self._parameters["Na_o"]
        F = self._parameters["F"]
        R = self._parameters["R"]
        T = self._parameters["T"]
        K_o = self._parameters["K_o"]

        # Init return args
        current = [ufl.zero()] * 1

        # Expressions for the Reversal potentials component
        E_Na = R * T * ufl.ln(Na_o / Na_i) / F
        E_K = R * T * ufl.ln(K_o / K_i) / F
        E_Ks = R * T * ufl.ln((Na_o * P_kna + K_o) / (K_i + P_kna * Na_i)) / F
        E_Ca = 0.5 * R * T * ufl.ln(Ca_o / Ca_i) / F

        # Expressions for the Inward rectifier potassium current component
        alpha_K1 = 0.1 / (1 + 6.14421235333e-06 * ufl.exp(-0.06 * E_K + 0.06 * V))
        beta_K1 = (
            3.06060402008 * ufl.exp(-0.0002 * E_K + 0.0002 * V)
            + 0.367879441171 * ufl.exp(-0.1 * E_K + 0.1 * V)
        ) / (1 + ufl.exp(-0.5 * V + 0.5 * E_K))
        xK1_inf = alpha_K1 / (beta_K1 + alpha_K1)
        i_K1 = 0.430331482912 * g_K1 * ufl.sqrt(K_o) * (-E_K + V) * xK1_inf

        # Expressions for the Rapid time dependent potassium current component
        i_Kr = 0.430331482912 * g_Kr * ufl.sqrt(K_o) * (-E_K + V) * Xr1 * Xr2

        # Expressions for the Slow time dependent potassium current component
        i_Ks = g_Ks * (Xs * Xs) * (-E_Ks + V)

        # Expressions for the Fast sodium current component
        i_Na = g_Na * (m * m * m) * (-E_Na + V) * h * j

        # Expressions for the Sodium background current component
        i_b_Na = g_bna * (-E_Na + V)

        # Expressions for the L_type Ca current component
        i_CaL = (
            4
            * g_CaL
            * (F * F)
            * (-15 + V)
            * (-Ca_o + 0.25 * Ca_ss * ufl.exp(F * (-30 + 2 * V) / (R * T)))
            * d
            * f
            * f2
            * fCass
            / (R * T * (-1 + ufl.exp(F * (-30 + 2 * V) / (R * T))))
        )

        # Expressions for the Calcium background current component
        i_b_Ca = g_bca * (-E_Ca + V)

        # Expressions for the Transient outward current component
        i_to = g_to * (-E_K + V) * r * s

        # Expressions for the Sodium potassium pump current component
        i_NaK = (
            K_o
            * P_NaK
            * Na_i
            / (
                (K_mNa + Na_i)
                * (K_mk + K_o)
                * (
                    1
                    + 0.1245 * ufl.exp(-0.1 * F * V / (R * T))
                    + 0.0353 * ufl.exp(-F * V / (R * T))
                )
            )
        )

        # Expressions for the Sodium calcium exchanger current component
        i_NaCa = (
            K_NaCa
            * (
                Ca_o * (Na_i * Na_i * Na_i) * ufl.exp(F * gamma * V / (R * T))
                - alpha
                * (Na_o * Na_o * Na_o)
                * Ca_i
                * ufl.exp(F * (-1 + gamma) * V / (R * T))
            )
            / (
                (1 + K_sat * ufl.exp(F * (-1 + gamma) * V / (R * T)))
                * (Ca_o + Km_Ca)
                * ((Km_Nai * Km_Nai * Km_Nai) + (Na_o * Na_o * Na_o))
            )
        )

        # Expressions for the Calcium pump current component
        i_p_Ca = g_pCa * Ca_i / (Ca_i + K_pCa)

        # Expressions for the Potassium pump current component
        i_p_K = g_pK * (-E_K + V) / (1 + 65.4052157419 * ufl.exp(-0.167224080268 * V))

        # Expressions for the Membrane component
        i_Stim = 0
        current[0] = (
            -i_b_Ca
            - i_K1
            - i_b_Na
            - i_to
            - i_p_Ca
            - i_Ks
            - i_p_K
            - i_NaK
            - i_Kr
            - i_Stim
            - i_NaCa
            - i_Na
            - i_CaL
        )

        # Return results
        return current[0]

    def I(self, v, s, time=None):
        """
        Transmembrane current

           I = -dV/dt

        """
        return -self._I(v, s, time)

    def F(self, v, s, time=None):
        """
        Right hand side for ODE system
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 18
        (
            Xr1,
            Xr2,
            Xs,
            m,
            h,
            j,
            d,
            f,
            f2,
            fCass,
            s,
            r,
            Ca_i,
            R_prime,
            Ca_SR,
            Ca_ss,
            Na_i,
            K_i,
        ) = s

        # Assign parameters
        P_kna = self._parameters["P_kna"]
        g_K1 = self._parameters["g_K1"]
        g_Kr = self._parameters["g_Kr"]
        g_Ks = self._parameters["g_Ks"]
        g_Na = self._parameters["g_Na"]
        g_bna = self._parameters["g_bna"]
        g_CaL = self._parameters["g_CaL"]
        g_bca = self._parameters["g_bca"]
        g_to = self._parameters["g_to"]
        K_mNa = self._parameters["K_mNa"]
        K_mk = self._parameters["K_mk"]
        P_NaK = self._parameters["P_NaK"]
        K_NaCa = self._parameters["K_NaCa"]
        K_sat = self._parameters["K_sat"]
        Km_Ca = self._parameters["Km_Ca"]
        Km_Nai = self._parameters["Km_Nai"]
        alpha = self._parameters["alpha"]
        gamma = self._parameters["gamma"]
        K_pCa = self._parameters["K_pCa"]
        g_pCa = self._parameters["g_pCa"]
        g_pK = self._parameters["g_pK"]
        Buf_c = self._parameters["Buf_c"]
        Buf_sr = self._parameters["Buf_sr"]
        Buf_ss = self._parameters["Buf_ss"]
        Ca_o = self._parameters["Ca_o"]
        EC = self._parameters["EC"]
        K_buf_c = self._parameters["K_buf_c"]
        K_buf_sr = self._parameters["K_buf_sr"]
        K_buf_ss = self._parameters["K_buf_ss"]
        K_up = self._parameters["K_up"]
        V_leak = self._parameters["V_leak"]
        V_rel = self._parameters["V_rel"]
        V_sr = self._parameters["V_sr"]
        V_ss = self._parameters["V_ss"]
        V_xfer = self._parameters["V_xfer"]
        Vmax_up = self._parameters["Vmax_up"]
        k1_prime = self._parameters["k1_prime"]
        k2_prime = self._parameters["k2_prime"]
        k3 = self._parameters["k3"]
        k4 = self._parameters["k4"]
        max_sr = self._parameters["max_sr"]
        min_sr = self._parameters["min_sr"]
        Na_o = self._parameters["Na_o"]
        Cm = self._parameters["Cm"]
        F = self._parameters["F"]
        R = self._parameters["R"]
        T = self._parameters["T"]
        V_c = self._parameters["V_c"]
        K_o = self._parameters["K_o"]

        # Init return args
        F_expressions = [ufl.zero()] * 18

        # Expressions for the Reversal potentials component
        E_Na = R * T * ufl.ln(Na_o / Na_i) / F
        E_K = R * T * ufl.ln(K_o / K_i) / F
        E_Ks = R * T * ufl.ln((Na_o * P_kna + K_o) / (K_i + P_kna * Na_i)) / F
        E_Ca = 0.5 * R * T * ufl.ln(Ca_o / Ca_i) / F

        # Expressions for the Inward rectifier potassium current component
        alpha_K1 = 0.1 / (1 + 6.14421235333e-06 * ufl.exp(-0.06 * E_K + 0.06 * V))
        beta_K1 = (
            3.06060402008 * ufl.exp(-0.0002 * E_K + 0.0002 * V)
            + 0.367879441171 * ufl.exp(-0.1 * E_K + 0.1 * V)
        ) / (1 + ufl.exp(-0.5 * V + 0.5 * E_K))
        xK1_inf = alpha_K1 / (beta_K1 + alpha_K1)
        i_K1 = 0.430331482912 * g_K1 * ufl.sqrt(K_o) * (-E_K + V) * xK1_inf

        # Expressions for the Rapid time dependent potassium current component
        i_Kr = 0.430331482912 * g_Kr * ufl.sqrt(K_o) * (-E_K + V) * Xr1 * Xr2

        # Expressions for the Xr1 gate component
        xr1_inf = 1.0 / (1 + ufl.exp(-26 / 7 - V / 7))
        alpha_xr1 = 450 / (1 + ufl.exp(-9 / 2 - V / 10))
        beta_xr1 = 6 / (1 + 13.5813245226 * ufl.exp(0.0869565217391 * V))
        tau_xr1 = alpha_xr1 * beta_xr1
        F_expressions[0] = (xr1_inf - Xr1) / tau_xr1

        # Expressions for the Xr2 gate component
        xr2_inf = 1.0 / (1 + ufl.exp(11 / 3 + V / 24))
        alpha_xr2 = 3 / (1 + ufl.exp(-3 - V / 20))
        beta_xr2 = 1.12 / (1 + ufl.exp(-3 + V / 20))
        tau_xr2 = alpha_xr2 * beta_xr2
        F_expressions[1] = (xr2_inf - Xr2) / tau_xr2

        # Expressions for the Slow time dependent potassium current component
        i_Ks = g_Ks * (Xs * Xs) * (-E_Ks + V)

        # Expressions for the Xs gate component
        xs_inf = 1.0 / (1 + ufl.exp(-5 / 14 - V / 14))
        alpha_xs = 1400 / ufl.sqrt(1 + ufl.exp(5 / 6 - V / 6))
        beta_xs = 1.0 / (1 + ufl.exp(-7 / 3 + V / 15))
        tau_xs = 80 + alpha_xs * beta_xs
        F_expressions[2] = (xs_inf - Xs) / tau_xs

        # Expressions for the Fast sodium current component
        i_Na = g_Na * (m * m * m) * (-E_Na + V) * h * j

        # Expressions for the m gate component
        m_inf = 1.0 / (
            (1 + 0.00184221158117 * ufl.exp(-0.110741971207 * V))
            * (1 + 0.00184221158117 * ufl.exp(-0.110741971207 * V))
        )
        alpha_m = 1.0 / (1 + ufl.exp(-12 - V / 5))
        beta_m = 0.1 / (1 + ufl.exp(-1 / 4 + V / 200)) + 0.1 / (1 + ufl.exp(7 + V / 5))
        tau_m = alpha_m * beta_m
        F_expressions[3] = (-m + m_inf) / tau_m

        # Expressions for the h gate component
        h_inf = 1.0 / (
            (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
            * (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
        )
        alpha_h = ufl.conditional(
            ufl.lt(V, -40), 4.43126792958e-07 * ufl.exp(-0.147058823529 * V), 0
        )
        beta_h = ufl.conditional(
            ufl.lt(V, -40),
            2.7 * ufl.exp(0.079 * V) + 310000 * ufl.exp(0.3485 * V),
            0.77 / (0.13 + 0.0497581410839 * ufl.exp(-0.0900900900901 * V)),
        )
        tau_h = 1.0 / (alpha_h + beta_h)
        F_expressions[4] = (-h + h_inf) / tau_h

        # Expressions for the j gate component
        j_inf = 1.0 / (
            (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
            * (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
        )
        alpha_j = ufl.conditional(
            ufl.lt(V, -40),
            (37.78 + V)
            * (-25428 * ufl.exp(0.2444 * V) - 6.948e-06 * ufl.exp(-0.04391 * V))
            / (1 + 50262745826.0 * ufl.exp(0.311 * V)),
            0,
        )
        beta_j = ufl.conditional(
            ufl.lt(V, -40),
            0.02424
            * ufl.exp(-0.01052 * V)
            / (1 + 0.0039608683399 * ufl.exp(-0.1378 * V)),
            0.6 * ufl.exp(0.057 * V) / (1 + 0.0407622039784 * ufl.exp(-0.1 * V)),
        )
        tau_j = 1.0 / (alpha_j + beta_j)
        F_expressions[5] = (-j + j_inf) / tau_j

        # Expressions for the Sodium background current component
        i_b_Na = g_bna * (-E_Na + V)

        # Expressions for the L_type Ca current component
        i_CaL = (
            4
            * g_CaL
            * (F * F)
            * (-15 + V)
            * (-Ca_o + 0.25 * Ca_ss * ufl.exp(F * (-30 + 2 * V) / (R * T)))
            * d
            * f
            * f2
            * fCass
            / (R * T * (-1 + ufl.exp(F * (-30 + 2 * V) / (R * T))))
        )

        # Expressions for the d gate component
        d_inf = 1.0 / (1 + 0.344153786865 * ufl.exp(-0.133333333333 * V))
        alpha_d = 0.25 + 1.4 / (1 + ufl.exp(-35 / 13 - V / 13))
        beta_d = 1.4 / (1 + ufl.exp(1 + V / 5))
        gamma_d = 1.0 / (1 + ufl.exp(5 / 2 - V / 20))
        tau_d = alpha_d * beta_d + gamma_d
        F_expressions[6] = (-d + d_inf) / tau_d

        # Expressions for the f gate component
        f_inf = 1.0 / (1 + ufl.exp(20 / 7 + V / 7))
        tau_f = (
            20
            + 1102.5 * ufl.exp(-((27 + V) * (27 + V)) / 225)
            + 180 / (1 + ufl.exp(3 + V / 10))
            + 200 / (1 + ufl.exp(13 / 10 - V / 10))
        )
        F_expressions[7] = (f_inf - f) / tau_f

        # Expressions for the F2 gate component
        f2_inf = 0.33 + 0.67 / (1 + ufl.exp(5 + V / 7))
        tau_f2 = (
            80 / (1 + ufl.exp(3 + V / 10))
            + 562 * ufl.exp(-((27 + V) * (27 + V)) / 240)
            + 31 / (1 + ufl.exp(5 / 2 - V / 10))
        )
        F_expressions[8] = (-f2 + f2_inf) / tau_f2

        # Expressions for the FCass gate component
        fCass_inf = 0.4 + 0.6 / (1 + 400.0 * (Ca_ss * Ca_ss))
        tau_fCass = 2 + 80 / (1 + 400.0 * (Ca_ss * Ca_ss))
        F_expressions[9] = (fCass_inf - fCass) / tau_fCass

        # Expressions for the Calcium background current component
        i_b_Ca = g_bca * (-E_Ca + V)

        # Expressions for the Transient outward current component
        i_to = g_to * (-E_K + V) * r * s

        # Expressions for the s gate component
        s_inf = 1.0 / (1 + ufl.exp(4 + V / 5))
        tau_s = (
            3
            + 5 / (1 + ufl.exp(-4 + V / 5))
            + 85 * ufl.exp(-((45 + V) * (45 + V)) / 320)
        )
        F_expressions[10] = (s_inf - s) / tau_s

        # Expressions for the r gate component
        r_inf = 1.0 / (1 + ufl.exp(10 / 3 - V / 6))
        tau_r = 0.8 + 9.5 * ufl.exp(-((40 + V) * (40 + V)) / 1800)
        F_expressions[11] = (r_inf - r) / tau_r

        # Expressions for the Sodium potassium pump current component
        i_NaK = (
            K_o
            * P_NaK
            * Na_i
            / (
                (K_mNa + Na_i)
                * (K_mk + K_o)
                * (
                    1
                    + 0.1245 * ufl.exp(-0.1 * F * V / (R * T))
                    + 0.0353 * ufl.exp(-F * V / (R * T))
                )
            )
        )

        # Expressions for the Sodium calcium exchanger current component
        i_NaCa = (
            K_NaCa
            * (
                Ca_o * (Na_i * Na_i * Na_i) * ufl.exp(F * gamma * V / (R * T))
                - alpha
                * (Na_o * Na_o * Na_o)
                * Ca_i
                * ufl.exp(F * (-1 + gamma) * V / (R * T))
            )
            / (
                (1 + K_sat * ufl.exp(F * (-1 + gamma) * V / (R * T)))
                * (Ca_o + Km_Ca)
                * ((Km_Nai * Km_Nai * Km_Nai) + (Na_o * Na_o * Na_o))
            )
        )

        # Expressions for the Calcium pump current component
        i_p_Ca = g_pCa * Ca_i / (Ca_i + K_pCa)

        # Expressions for the Potassium pump current component
        i_p_K = g_pK * (-E_K + V) / (1 + 65.4052157419 * ufl.exp(-0.167224080268 * V))

        # Expressions for the Calcium dynamics component
        i_up = Vmax_up / (1 + (K_up * K_up) / (Ca_i * Ca_i))
        i_leak = V_leak * (Ca_SR - Ca_i)
        i_xfer = V_xfer * (Ca_ss - Ca_i)
        kcasr = max_sr - (max_sr - min_sr) / (1 + (EC * EC) / (Ca_SR * Ca_SR))
        Ca_i_bufc = 1.0 / (1 + Buf_c * K_buf_c / ((Ca_i + K_buf_c) * (Ca_i + K_buf_c)))
        Ca_sr_bufsr = 1.0 / (
            1 + Buf_sr * K_buf_sr / ((Ca_SR + K_buf_sr) * (Ca_SR + K_buf_sr))
        )
        Ca_ss_bufss = 1.0 / (
            1 + Buf_ss * K_buf_ss / ((K_buf_ss + Ca_ss) * (K_buf_ss + Ca_ss))
        )
        F_expressions[12] = (
            -Cm * (i_b_Ca - 2 * i_NaCa + i_p_Ca) / (2 * F * V_c)
            + i_xfer
            + V_sr * (-i_up + i_leak) / V_c
        ) * Ca_i_bufc
        k1 = k1_prime / kcasr
        k2 = k2_prime * kcasr
        O = (Ca_ss * Ca_ss) * R_prime * k1 / (k3 + (Ca_ss * Ca_ss) * k1)
        F_expressions[13] = -Ca_ss * R_prime * k2 + k4 * (1 - R_prime)
        i_rel = V_rel * (Ca_SR - Ca_ss) * O
        F_expressions[14] = (i_up - i_leak - i_rel) * Ca_sr_bufsr
        F_expressions[15] = (
            V_sr * i_rel / V_ss - Cm * i_CaL / (2 * F * V_ss) - V_c * i_xfer / V_ss
        ) * Ca_ss_bufss

        # Expressions for the Sodium dynamics component
        F_expressions[16] = Cm * (-i_b_Na - 3 * i_NaCa - 3 * i_NaK - i_Na) / (F * V_c)

        # Expressions for the Membrane component
        i_Stim = 0

        # Expressions for the Potassium dynamics component
        F_expressions[17] = (
            Cm * (-i_K1 + 2 * i_NaK - i_to - i_Ks - i_p_K - i_Kr - i_Stim) / (F * V_c)
        )

        # Return results
        return dolfin.as_vector(F_expressions)

    def num_states(self):
        return 18

    def __str__(self):
        return "Tentusscher_panfilov_2006_epi_cell cardiac cell model"



"""
This module contains splitting solvers for CardiacModel objects. In
particular, the classes

  * SplittingSolver
  * BasicSplittingSolver

These solvers solve the bidomain (or monodomain) equations on the
form: find the transmembrane potential :math:`v = v(x, t)` in mV, the
extracellular potential :math:`u = u(x, t)` in mV, and any additional
state variables :math:`s = s(x, t)` such that

.. math::

   v_t - \mathrm{div} (M_i \mathrm{grad} v + M_i \mathrm{grad} u) = - I_{ion}(v, s) + I_s

         \mathrm{div} (M_i \mathrm{grad} v + (M_i + M_e) \mathrm{grad} u) = I_a

   s_t = F(v, s)

where

  * the subscript :math:`t` denotes the time derivative,
  * :math:`M_i` and :math:`M_e` are conductivity tensors (in mm^2/ms)
  * :math:`I_s` is prescribed input current (in mV/ms)
  * :math:`I_a` is prescribed input current (in mV/ms)
  * :math:`I_{ion}` and :math:`F` are typically specified by a cell model

Note that M_i and M_e can be viewed as scaled by :math:`\chi*C_m` where
  * :math:`\chi` is the surface-to volume ratio of cells (in 1/mm) ,
  * :math:`C_m` is the specific membrane capacitance (in mu F/(mm^2) ),

In addition, initial conditions are given for :math:`v` and :math:`s`:

.. math::

   v(x, 0) = v_0

   s(x, 0) = s_0

Finally, boundary conditions must be prescribed. These solvers assume
pure Neumann boundary conditions for :math:`v` and :math:`u` and
enforce the additional average value zero constraint for u.

The solvers take as input a
:py:class:`~cbcbeat.cardiacmodels.CardiacModel` providing the
required input specification of the problem. In particular, the
applied current :math:`I_a` is extracted from the
:py:attr:`~cbcbeat.cardiacmodels.CardiacModel.applied_current`
attribute, while the stimulus :math:`I_s` is extracted from the
:py:attr:`~cbcbeat.cardiacmodels.CardiacModel.stimulus` attribute.

It should be possible to use the solvers interchangably. However, note
that the BasicSplittingSolver is not optimised and should be used for
testing or debugging purposes primarily.
"""

# Copyright (C) 2012-2013 Marie E. Rognes (meg@simula.no)
# Use and modify at will
# Last changed: 2013-04-15

__all__ = [
    "SplittingSolver",
    "BasicSplittingSolver",
]

import dolfin
from cbcbeat.dolfinimport import backend

from cbcbeat import CardiacModel
from cbcbeat.cellsolver import BasicCardiacODESolver, CardiacODESolver
from cbcbeat.bidomainsolver import BasicBidomainSolver, BidomainSolver
from cbcbeat.monodomainsolver import BasicMonodomainSolver, MonodomainSolver
from cbcbeat.utils import TimeStepper, annotate_kwargs
from ufl.log import info_blue

progress = dolfin.LogLevel.PROGRESS


class Interpolator:
    """A simple helper class to mimic the API of a FunctionAssigner
    assuming that it is possible to interpolate between the two
    functions passed in to the `assign` method
    """

    def assign(self, v_ode, v, **kwargs):
        v_ode.interpolate(v)


def setup_merger(ode_space, vur_space):
    ode_el = ode_space.ufl_element()
    vur_el = vur_space.ufl_element()

    if ode_el.family() == vur_el.family():
        return backend.FunctionAssigner(ode_space, vur_space)

    return Interpolator()


class BasicSplittingSolver:
    """

    A non-optimised solver for the bidomain equations based on the
    operator splitting scheme described in Sundnes et al 2006, p. 78
    ff.

    The solver computes as solutions:

      * "vs" (:py:class:`dolfin.Function`) representing the solution
        for the transmembrane potential and any additional state
        variables, and
      * "vur" (:py:class:`dolfin.Function`) representing the
        transmembrane potential in combination with the extracellular
        potential and an additional Lagrange multiplier.

    The algorithm can be controlled by a number of parameters. In
    particular, the splitting algorithm can be controlled by the
    parameter "theta": "theta" set to 1.0 corresponds to a (1st order)
    Godunov splitting while "theta" set to 0.5 to a (2nd order) Strang
    splitting.

    This solver has not been optimised for computational efficiency
    and should therefore primarily be used for debugging purposes. For
    an equivalent, but more efficient, solver, see
    :py:class:`cbcbeat.splittingsolver.SplittingSolver`.

    *Arguments*
      model (:py:class:`cbcbeat.cardiacmodels.CardiacModel`)
        a CardiacModel object describing the simulation set-up
      params (:py:class:`dolfin.Parameters`, optional)
        a Parameters object controlling solver parameters
      V_index (int)
        Index for the membrane potential among the state
        variables, by default 0

    *Assumptions*
      * The cardiac conductivities do not vary in time

    """

    def __init__(self, model, params=None, V_index=0):
        "Create solver from given Cardiac Model and (optional) parameters."

        assert isinstance(
            model, CardiacModel
        ), "Expecting CardiacModel as first argument"

        # Set model and parameters
        self._model = model
        self._V_index = V_index
        self.parameters = self.default_parameters()
        if params is not None:
            self.parameters.update(params)

        # Extract solution domain
        self._domain = self._model.domain()
        self._time = self._model.time()

        # Create ODE solver and extract solution fields
        self.ode_solver = self._create_ode_solver()
        (self.vs_, self.vs) = self.ode_solver.solution_fields()
        self.VS = self.vs.function_space()

        # Create PDE solver and extract solution fields
        self.pde_solver = self._create_pde_solver()
        (self.v_, self.vur) = self.pde_solver.solution_fields()

        # Create function assigner for merging v from self.vur into self.vs[0]
        if self.parameters["pde_solver"] == "bidomain":
            V = self.vur.function_space().sub(0)
        else:
            V = self.vur.function_space()

        self.merger = setup_merger(self._ode_membrane_potential_space, V)

        self._annotate_kwargs = annotate_kwargs(self.parameters)

    def _create_ode_solver(self):
        """Helper function to initialize a suitable ODE solver from
        the cardiac model."""

        # Extract cardiac cell model from cardiac model
        cell_model = self._model.cell_models()

        # Extract stimulus from the cardiac model(!)
        if self.parameters["apply_stimulus_current_to_pde"]:
            stimulus = None
        else:
            stimulus = self._model.stimulus()

        # Extract ode solver parameters
        params = self.parameters["BasicCardiacODESolver"]
        # Propagate enable_adjoint to Bidomain solver
        if params.has_parameter("enable_adjoint"):
            params["enable_adjoint"] = self.parameters["enable_adjoint"]

        solver = BasicCardiacODESolver(
            self._domain, self._time, cell_model, I_s=stimulus, params=params
        )
        return solver

    def _create_pde_solver(self):
        """Helper function to initialize a suitable PDE solver from
        the cardiac model."""

        # Extract applied current from the cardiac model
        applied_current = self._model.applied_current()

        # Extract stimulus from the cardiac model if we should apply
        # it to the PDEs (in the other case, it is handled by the ODE
        # solver)
        if self.parameters["apply_stimulus_current_to_pde"]:
            stimulus = self._model.stimulus()
        else:
            stimulus = None

        # Extract conductivities from the cardiac model
        (M_i, M_e) = self._model.conductivities()

        if self.parameters["pde_solver"] == "bidomain":
            PDESolver = BasicBidomainSolver
            params = self.parameters["BasicBidomainSolver"]
            args = (self._domain, self._time, M_i, M_e)
            kwargs = dict(
                I_s=stimulus,
                I_a=applied_current,
                v_=self._ode_membrane_potential_func,
                params=params,
            )
        else:
            PDESolver = BasicMonodomainSolver
            params = self.parameters["BasicMonodomainSolver"]
            args = (self._domain, self._time, M_i)
            kwargs = dict(
                I_s=stimulus, v_=self._ode_membrane_potential_func, params=params
            )

        # Propagate enable_adjoint to Bidomain solver
        if params.has_parameter("enable_adjoint"):
            params["enable_adjoint"] = self.parameters["enable_adjoint"]

        solver = PDESolver(*args, **kwargs)

        return solver

    @property
    def _ode_membrane_potential_space(self):
        """Get the functio space for the membrane potential
        in the ODE
        """
        if self.VS.num_sub_spaces() == 0:
            return self.VS
        return self.VS.sub(self._V_index)

    def _extract_V(self, solution):
        """Extract the membrane potential from the solution"""
        if self.VS.num_sub_spaces() == 0:
            return solution
        return solution.sub(self._V_index)

    @property
    def _ode_membrane_potential_func(self):
        """Get the membrane potential from the ODE"""
        if self.vs.value_rank() == 0:
            # The solution from the ODE does not come from a
            # VectorFunctionSpace. Assme that this is the membrane
            # potential
            return self.vs
        else:
            # We retrieve the membrane potential from the
            # correct subspace
            return self.vs[self._V_index]

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters for the
        splitting solver

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)

        To inspect all the default parameters, do::

          info(BasicSplittingSolver.default_parameters(), True)

        """

        params = dolfin.Parameters("BasicSplittingSolver")
        params.add("enable_adjoint", True)
        params.add("theta", 0.5, 0.0, 1.0)
        params.add("apply_stimulus_current_to_pde", False)
        try:
            params.add("pde_solver", "bidomain", set(["bidomain", "monodomain"]))
        except Exception:
            params.add("pde_solver", "bidomain", ["bidomain", "monodomain"])

        # Add default parameters from ODE solver, but update for V
        # space
        ode_solver_params = BasicCardiacODESolver.default_parameters()
        ode_solver_params["V_polynomial_degree"] = 1
        ode_solver_params["V_polynomial_family"] = "CG"
        params.add(ode_solver_params)

        pde_solver_params = BasicBidomainSolver.default_parameters()
        pde_solver_params["polynomial_degree"] = 1
        params.add(pde_solver_params)

        pde_solver_params = BasicMonodomainSolver.default_parameters()
        pde_solver_params["polynomial_degree"] = 1
        params.add(pde_solver_params)

        return params

    def solution_fields(self):
        """
        Return tuple of previous and current solution objects.

        Modifying these will modify the solution objects of the solver
        and thus provides a way for setting initial conditions for
        instance.

        *Returns*
          (previous vs, current vs, current vur) (:py:class:`tuple` of :py:class:`dolfin.Function`)
        """
        return (self.vs_, self.vs, self.vur)

    def solve(self, interval, dt):
        """
        Solve the problem given by the model on a given time interval
        (t0, t1) with a given timestep dt and return generator for a
        tuple of the time step and the solution fields.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval for the solve given by (t0, t1)
          dt (int, list of tuples of floats)
            The timestep for the solve. A list of tuples of floats can
            also be passed. Each tuple should contain two floats where the
            first includes the start time and the second the dt.

        *Returns*
          (timestep, solution_fields) via (:py:class:`genexpr`)

        *Example of usage*::

          # Create generator
          dts = [(0., 0.1), (1.0, 0.05), (2.0, 0.1)]
          solutions = solver.solve((0.0, 1.0), dts)

          # Iterate over generator (computes solutions as you go)
          for ((t0, t1), (vs_, vs, vur)) in solutions:
            # do something with the solutions

        """

        # Create timestepper
        time_stepper = TimeStepper(
            interval, dt, annotate=self.parameters["enable_adjoint"]
        )

        for t0, t1 in time_stepper:
            info_blue("Solving on t = (%g, %g)" % (t0, t1))
            self.step((t0, t1))

            # Yield solutions
            yield (t0, t1), self.solution_fields()

            # Update previous solution
            self.vs_.assign(self.vs)

    def step(self, interval):
        """
        Solve the problem given by the model on a given time interval
        (t0, t1) with timestep given by the interval length.

        *Arguments*
          interval (:py:class:`tuple`)
            The time interval for the solve given by (t0, t1)

        *Invariants*
          Given self._vs in a correct state at t0, provide v and s (in
          self.vs) and u (in self.vur) in a correct state at t1. (Note
          that self.vur[0] == self.vs[0] only if theta = 1.0.)
        """

        # Extract some parameters for readability
        theta = self.parameters["theta"]

        # Extract time domain
        (t0, t1) = interval
        dt = t1 - t0
        t = t0 + theta * dt

        # Compute tentative membrane potential and state (vs_star)
        dolfin.begin(progress, "Tentative ODE step")
        # Assumes that its vs_ is in the correct state, gives its vs
        # in the current state
        self.ode_solver.step((t0, t))
        dolfin.end()

        # Compute tentative potentials vu = (v, u)
        dolfin.begin(progress, "PDE step")
        # Assumes that its vs_ is in the correct state, gives vur in
        # the current state
        self.pde_solver.step((t0, t1))
        dolfin.end()

        # If first order splitting, we need to ensure that self.vs is
        # up to date, but otherwise we are done.
        if theta == 1.0:
            # Assumes that the v part of its vur and the s part of its
            # vs are in the correct state, provides input argument(in
            # this case self.vs) in its correct state
            self.merge(self.vs)
            return

        # Otherwise, we do another ode_step:
        dolfin.begin(progress, "Corrective ODE step")

        # Update vs_ based on vs, the s part of vs and vs_ are now in
        # the correct state.
        self.vs_.assign(self.vs)

        # Set the v part of vs_ to be v from vur. Now vs_ is in its
        # correct state.
        self.merge(self.vs_)

        # Assumes that its vs_ is in the correct state, provides vs in
        # the correct state
        self.ode_solver.step((t, t1))

        dolfin.end()

    def merge(self, solution):
        """
        Combine solutions from the PDE solve and the ODE solve to form
        a single mixed function.

        *Arguments*
          solution (:py:class:`dolfin.Function`)
            Function holding the combined result
        """
        timer = dolfin.Timer("Merge step")

        dolfin.begin(progress, "Merging")
        if self.parameters["pde_solver"] == "bidomain":
            v = self.vur.sub(0)
        else:
            v = self.vur

        self.merger.assign(self._extract_V(solution), v, **self._annotate_kwargs)
        dolfin.end()

        timer.stop()


class SplittingSolver(BasicSplittingSolver):
    """

    An optimised solver for the bidomain equations based on the
    operator splitting scheme described in Sundnes et al 2006, p. 78
    ff.

    The solver computes as solutions:

      * "vs" (:py:class:`dolfin.Function`) representing the solution
        for the transmembrane potential and any additional state
        variables, and
      * "vur" (:py:class:`dolfin.Function`) representing the
        transmembrane potential in combination with the extracellular
        potential and an additional Lagrange multiplier.

    The algorithm can be controlled by a number of parameters. In
    particular, the splitting algorithm can be controlled by the
    parameter "theta": "theta" set to 1.0 corresponds to a (1st order)
    Godunov splitting while "theta" set to 0.5 to a (2nd order) Strang
    splitting.

    *Arguments*
      model (:py:class:`cbcbeat.cardiacmodels.CardiacModel`)
        a CardiacModel object describing the simulation set-up
      params (:py:class:`dolfin.Parameters`, optional)
        a Parameters object controlling solver parameters

    *Example of usage*::

      from cbcbeat import *

      # Describe the cardiac model
      mesh = UnitSquareMesh(100, 100)
      time = Constant(0.0)
      cell_model = FitzHughNagumoManual()
      stimulus = Expression("10*t*x[0]", t=time, degree=1)
      cm = CardiacModel(mesh, time, 1.0, 1.0, cell_model, stimulus)

      # Extract default solver parameters
      ps = SplittingSolver.default_parameters()

      # Examine the default parameters
      info(ps, True)

      # Update parameter dictionary
      ps["theta"] = 1.0 # Use first order splitting
      ps["CardiacODESolver"]["scheme"] = "GRL1" # Use Generalized Rush-Larsen scheme

      # Use monodomain equations as the PDE model
      ps["pde_solver"] = "monodomain"
      # Use direct linear solver of the PDEs
      ps["MonodomainSolver"]["linear_solver_type"] = "direct"
      # Use backward Euler for temporal discretization for the PDEs
      ps["MonodomainSolver"]["theta"] = 1.0

      solver = SplittingSolver(cm, params=ps)

      # Extract the solution fields and set the initial conditions
      (vs_, vs, vur) = solver.solution_fields()
      vs_.assign(cell_model.initial_conditions())

      # Solve
      dt = 0.1
      T = 1.0
      interval = (0.0, T)
      for (timestep, fields) in solver.solve(interval, dt):
          (vs_, vs, vur) = fields
          # Do something with the solutions


    *Assumptions*
      * The cardiac conductivities do not vary in time

    """

    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters for the
        splitting solver

        *Returns*
          The set of default parameters (:py:class:`dolfin.Parameters`)

        *Example of usage*::

          info(SplittingSolver.default_parameters(), True)
        """

        params = dolfin.Parameters("SplittingSolver")
        params.add("enable_adjoint", True)
        params.add("theta", 0.5, 0, 1)
        params.add("apply_stimulus_current_to_pde", False)
        try:
            params.add("pde_solver", "bidomain", set(["bidomain", "monodomain"]))
            params.add(
                "ode_solver_choice",
                "CardiacODESolver",
                set(["BasicCardiacODESolver", "CardiacODESolver"]),
            )
        except Exception:
            params.add("pde_solver", "bidomain", ["bidomain", "monodomain"])
            params.add(
                "ode_solver_choice",
                "CardiacODESolver",
                ["BasicCardiacODESolver", "CardiacODESolver"],
            )
            pass

        # Add default parameters from ODE solver
        ode_solver_params = CardiacODESolver.default_parameters()
        ode_solver_params["scheme"] = "RL1"
        params.add(ode_solver_params)

        # Add default parameters from ODE solver
        basic_ode_solver_params = BasicCardiacODESolver.default_parameters()
        basic_ode_solver_params["V_polynomial_degree"] = 1
        basic_ode_solver_params["V_polynomial_family"] = "CG"
        params.add(basic_ode_solver_params)

        pde_solver_params = BidomainSolver.default_parameters()
        pde_solver_params["polynomial_degree"] = 1
        params.add(pde_solver_params)

        pde_solver_params = MonodomainSolver.default_parameters()
        pde_solver_params["polynomial_degree"] = 1
        params.add(pde_solver_params)

        return params

    def _create_ode_solver(self):
        """Helper function to initialize a suitable ODE solver from
        the cardiac model."""

        # Extract cardiac cell model from cardiac model
        cell_model = self._model.cell_models()

        # Extract stimulus from the cardiac model(!)
        if self.parameters["apply_stimulus_current_to_pde"]:
            stimulus = None
        else:
            stimulus = self._model.stimulus()

        Solver = eval(self.parameters["ode_solver_choice"])
        params = self.parameters[Solver.__name__]
        if params.has_parameter("enable_adjoint"):
            params["enable_adjoint"] = self.parameters["enable_adjoint"]
        solver = Solver(
            self._domain, self._time, cell_model, I_s=stimulus, params=params
        )

        return solver

    def _create_pde_solver(self):
        """Helper function to initialize a suitable PDE solver from
        the cardiac model."""

        # Extract applied current from the cardiac model (stimulus
        # invoked in the ODE step)
        applied_current = self._model.applied_current()

        # Extract stimulus from the cardiac model
        if self.parameters["apply_stimulus_current_to_pde"]:
            stimulus = self._model.stimulus()
        else:
            stimulus = None

        # Extract conductivities from the cardiac model
        (M_i, M_e) = self._model.conductivities()

        if self.parameters["pde_solver"] == "bidomain":
            PDESolver = BidomainSolver
            params = self.parameters["BidomainSolver"]
            args = (self._domain, self._time, M_i, M_e)
            kwargs = dict(
                I_s=stimulus,
                I_a=applied_current,
                v_=self._ode_membrane_potential_func,
                params=params,
            )
        else:
            PDESolver = MonodomainSolver
            params = self.parameters["MonodomainSolver"]
            args = (self._domain, self._time, M_i)
            kwargs = dict(
                I_s=stimulus, v_=self._ode_membrane_potential_func, params=params
            )

        # Propagate enable_adjoint to Bidomain solver
        if params.has_parameter("enable_adjoint"):
            params["enable_adjoint"] = self.parameters["enable_adjoint"]

        solver = PDESolver(*args, **kwargs)

        return solver



"""
The cbcbeat Python module is a problem and solver collection for
cardiac electrophysiology models.

To import the module, type::

  from cbcbeat import *

"""
import warnings as _warnings


# Import all of dolfin with possibly dolfin-adjoint on top


# Model imports
from cbcbeat.cellmodels import (
    FitzHughNagumoManual,
    NoCellModel,
    RogersMcCulloch,
    Beeler_reuter_1977,
    Tentusscher_2004_mcell,
    Tentusscher_panfilov_2006_epi_cell,
    Tentusscher_panfilov_2006_M_cell,
    Fenton_karma_1998_MLR1_altered,
    Fenton_karma_1998_BR_altered,
    CardiacCellModel,
    MultiCellModel,
    supported_cell_models,
)
from cbcbeat.markerwisefield import (
    rhs_with_markerwise_field,
    Markerwise,
    handle_markerwise,
)

from cbcbeat.bidomainsolver import BasicBidomainSolver, BidomainSolver
from cbcbeat.cardiacmodels import CardiacModel
from cbcbeat.cellsolver import (
    CardiacODESolver,
    BasicCardiacODESolver,
    SingleCellSolver,
    BasicSingleCellSolver,
)
from cbcbeat.monodomainsolver import BasicMonodomainSolver, MonodomainSolver
from cbcbeat.splittingsolver import SplittingSolver, BasicSplittingSolver
from cbcbeat.dolfinimport import backend
from cbcbeat import dolfinimport

# Solver imports

# Various utility functions, mainly for internal use

# NB: Workaround for FEniCS 1.7.0dev
import ufl as _ufl

_ufl.algorithms.apply_derivatives.CONDITIONAL_WORKAROUND = True
import dolfin as _dolfin

parameters = _dolfin.parameters
# Set-up some global parameters
beat_parameters = _dolfin.Parameters("beat-parameters")
beat_parameters.add("enable_adjoint", True)

import ufl
import ufl.log
from ufl import *
from ufl.log import *
from dolfin import *


try:
    import dolfin_adjoint
    from dolfin_adjoint import *
except ImportError:
    dolfin_adjoint = None
    dolfin_adjoint_all = []
else:
    dolfin_adjoint_all = dolfin_adjoint.__all__


class StarImportDeprecationWarning(DeprecationWarning):
    def __init__(self, *args: object) -> None:
        msg = """
*** ===================================================== ***
*** CBCBeat: Using * imports will be deprecated in the    ***
*** next version of cbcbeat. Please change all imports    ***
*** such as 'from cbcbeat import *' with 'import cbcbeat' ***
*** or 'from cbcbeat import foo, bar'                     ***
*** ===================================================== ***"""
        super().__init__(msg)


_warnings.simplefilter("always", StarImportDeprecationWarning)


def __getattr__(name):
    if name == "__all__":
        _warnings.warn(StarImportDeprecationWarning(), stacklevel=2)
        return (
            dolfin_adjoint_all
            + _dolfin.__all__
            + ufl.__all__
            + ufl.log.__all__
            + [
                "FitzHughNagumoManual",
                "NoCellModel",
                "RogersMcCulloch",
                "Beeler_reuter_1977",
                "Tentusscher_2004_mcell",
                "Tentusscher_panfilov_2006_epi_cell",
                "Tentusscher_panfilov_2006_M_cell",
                "Fenton_karma_1998_MLR1_altered",
                "Fenton_karma_1998_BR_altered",
                "CardiacCellModel",
                "MultiCellModel",
                "rhs_with_markerwise_field",
                "Markerwise",
                "handle_markerwise",
                "BasicBidomainSolver",
                "BidomainSolver",
                "CardiacModel",
                "CardiacODESolver",
                "BasicCardiacODESolver",
                "SingleCellSolver",
                "BasicSingleCellSolver",
                "BasicMonodomainSolver",
                "MonodomainSolver",
                "SplittingSolver",
                "BasicSplittingSolver",
                "backend",
                "supported_cell_models",
                "dolfinimport",
                "parameters",
                "beat_parameters",
            ]
        )

    msg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(msg)



"""This module provides various utilities for internal use."""

__author__ = "Marie E. Rognes (meg@simula.no), 2012--2013"

__all__ = ["state_space", "end_of_time", "convergence_rate", "Projecter"]

import math
import dolfin
from cbcbeat.dolfinimport import backend, has_dolfin_adjoint
from dolfin.cpp.log import log, LogLevel


def annotate_kwargs(ba_parameters):
    if not has_dolfin_adjoint:
        return {}
    if not ba_parameters["enable_adjoint"]:
        return {"annotate": False}
    if dolfin.parameters["adjoint"]["stop_annotating"]:
        return {"annotate": False}

    return {"annotate": True}


def splat(vs, dim):
    if vs.function_space().ufl_element().num_sub_elements() == dim:
        v = vs[0]
        if dim == 2:
            s = vs[1]
        else:
            s = dolfin.as_vector([vs[i] for i in range(1, dim)])
    else:
        v, s = dolfin.split(vs)

    return v, s


def state_space(domain, d, family=None, k=1):
    """Return function space for the state variables.

    *Arguments*
      domain (:py:class:`dolfin.Mesh`)
        The computational domain
      d (int)
        The number of states
      family (string, optional)
        The finite element family, defaults to "CG" if None is given.
      k (int, optional)
        The finite element degree, defaults to 1

    *Returns*
      a function space (:py:class:`dolfin.FunctionSpace`)
    """
    if family is None:
        family = "CG"
    if d > 1:
        S = dolfin.VectorFunctionSpace(domain, family, k, d)
    else:
        S = dolfin.FunctionSpace(domain, family, k)
    return S


def end_of_time(T, t0, t1, dt):
    """
    Return True if the interval (t0, t1) is the last before the end
    time T, otherwise False.
    """
    return (t1 + dt) > (T + dolfin.DOLFIN_EPS)


class TimeStepper:
    """
    A helper object that keep track of simulated time
    """

    def __init__(self, interval, dt, annotate=False):
        """
        *Arguments*
          interval (:py:class:`tuple`)
            The time interval for the solve given by (t0, t1)
          dt (:py:class:`int or :py:class:`list` of py:class:`tuples` of py:class:`float`)
            The timestep for the solve. A list of tuples of floats can
            also be passed. Each tuple should contain two floats where the
            first includes the start time and the second the dt.
          annotate (:py:class:`bool)
            If enabling dolfin_adjoint timestep annotation
        """

        self.annotate = annotate

        if (
            not isinstance(interval, (tuple, list))
            or len(interval) != 2
            or not all(isinstance(value, (float, int)) for value in interval)
        ):
            raise TypeError(
                "expected tuple or list of size 2 with scalars for "
                "the interval argument"
            )

        if interval[0] >= interval[1]:
            raise ValueError(
                "Start time need to be larger than stop time: "
                "interval[0] < interval[1]"
            )

        # Store time interval
        (self.T0, self.T1) = interval

        if not isinstance(dt, (float, int, list)):
            raise TypeError("expected float or list of tuples for dt argument")

        if isinstance(dt, (float, int)):
            dt = [(self.T0, dt)]

        # Check that all dt are tuples of size 2 with either floats or ints.
        if any(
            (
                not isinstance(item, tuple)
                or len(item) != 2
                or not all(isinstance(value, (float, int)) for value in item)
            )
            for item in dt
        ):
            raise TypeError(
                "expected list of tuples of size 2 with scalars for " "the dt argument"
            )

        # Check that first time value of dt is the same as the first given in interval
        if dt[0][0] != self.T0:
            raise ValueError(
                "expected first time value of dt to be the same as "
                "the first value of time interval."
            )

        # Check that all time values given in dt are increasing
        if not all(dt[i][0] < dt[i + 1][0] for i in range(len(dt) - 1)):
            raise ValueError("expected all time values in dt to be increasing")

        # Check that all time step values given in dt are positive
        if not all(dt[i][1] > 0 for i in range(len(dt))):
            raise ValueError("expected all time step values in dt to be positive")

        # Store dt
        self._dt = dt

        # Add a dummy dt including stop interval time
        if self._dt[-1][0] < self.T1:
            self._dt.append((self.T1, self._dt[-1][1]))

        # Keep track of dt index
        self._dt_ind = 0
        self.t0 = self.T0
        # self.t1 = self.T0 + self.dt

        # Step through time steps until at end time.
        if self.annotate and has_dolfin_adjoint:
            backend.adj_start_timestep(self.T0)

    def __iter__(self):
        """
        Return an iterator over time intervals
        """
        eps = 1e-10

        while True:
            # Get next t1
            t1 = self.next_t1()

            # Yield time interval
            yield self.t0, t1

            # Break if this is the last step
            if abs(t1 - self.T1) < eps:
                if self.annotate and has_dolfin_adjoint:
                    backend.adj_inc_timestep(time=t1, finished=True)
                break

            # Move to next time
            if self.annotate and has_dolfin_adjoint:
                backend.adj_inc_timestep(time=t1)

            self.t0 = t1

    def next_t1(self):
        """
        Return the time of next end interval
        """
        assert self._dt_ind < len(self._dt) + 1
        dt = self._dt[self._dt_ind][1]
        time_to_switch_dt = self._dt[self._dt_ind + 1][0]
        if time_to_switch_dt - dolfin.DOLFIN_EPS > self.t0 + dt:
            return self.t0 + dt

        # Update dt index
        self._dt_ind += 1
        return time_to_switch_dt


def convergence_rate(hs, errors):
    """
    Compute and return rates of convergence :math:`r_i` such that

    .. math::

      errors = C hs^r

    """
    assert len(hs) == len(errors), "hs and errors must have same length."
    # Compute converence rates
    rates = [
        (math.log(errors[i + 1] / errors[i])) / (math.log(hs[i + 1] / hs[i]))
        for i in range(len(hs) - 1)
    ]

    # Return convergence rates
    return rates


class Projecter(object):
    """Customized class for repeated projection.

    *Arguments*
      V (:py:class:`dolfin.FunctionSpace`)
        The function space to project into
      solver_type (string, optional)
        "iterative" (default) or "direct"

    *Example of usage*::
      my_project = Projecter(V, solver_type="direct")
      u = Function(V)
      f = Function(W)
      my_project(f, u)
    """

    def __init__(self, V, params=None):
        # Set parameters
        self.parameters = self.default_parameters()
        if params is not None:
            self.parameters.update(params)

        # Set-up mass matrix for L^2 projection
        self.V = V
        self.u = dolfin.TrialFunction(self.V)
        self.v = dolfin.TestFunction(self.V)
        self.m = dolfin.inner(self.u, self.v) * dolfin.dx()
        self.M = backend.assemble(self.m)
        self.b = dolfin.Vector(V.mesh().mpi_comm(), V.dim())

        solver_type = self.parameters["linear_solver_type"]
        assert (
            solver_type == "lu" or solver_type == "cg"
        ), "Expecting 'linear_solver_type' to be 'lu' or 'cg'"
        if solver_type == "lu":
            log(LogLevel.TRACE, "Setting up direct solver for projecter")

            # Customize LU solver (reuse everything)
            solver = backend.LUSolver(self.M)
        else:
            log(LogLevel.TRACE, "Setting up iterative solver for projecter")
            # Customize Krylov solver (reuse everything)
            solver = backend.KrylovSolver("cg", "ilu")
            solver.set_operator(self.M)
            if solver.parameters.has_parameter("preconditioner"):
                solver.parameters["preconditioner"]["structure"] = "same"
            # solver.parameters["nonzero_initial_guess"] = True
        self.solver = solver

    @staticmethod
    def default_parameters():
        parameters = dolfin.Parameters("Projecter")
        parameters.add("linear_solver_type", "cg")
        return parameters

    def __call__(self, f, u):
        """
        Carry out projection of ufl Expression f and store result in
        the function u. The user must make sure that u lives in the
        right space.

        *Arguments*
          f (:py:class:`ufl.Expr`)
            The thing to be projected into this function space
          u (:py:class:`dolfin.Function`)
            The result of the projection
        """
        L = dolfin.inner(f, self.v) * dolfin.dx()
        backend.assemble(L, tensor=self.b)
        self.solver.solve(u.vector(), self.b)



"""This module contains a base class for cardiac cell models."""
__author__ = "Marie E. Rognes (meg@simula.no), 2012--2013"
__all__ = ["CardiacCellModel", "MultiCellModel"]

from ufl.log import error
import dolfin
from cbcbeat.dolfinimport import backend
from collections import OrderedDict


class CardiacCellModel:
    """
    Base class for cardiac cell models. Specialized cell models should
    subclass this class.

    Essentially, a cell model represents a system of ordinary
    differential equations. A cell model is here described by two
    (Python) functions, named F and I. The model describes the
    behavior of the transmembrane potential 'v' and a number of state
    variables 's'

    The function F gives the right-hand side for the evolution of the
    state variables:

      d/dt s = F(v, s)

    The function I gives the ionic current. If a single cell is
    considered, I gives the (negative) right-hand side for the
    evolution of the transmembrane potential

    (*)  d/dt v = - I(v, s)

    If used in a bidomain setting, the ionic current I enters into the
    parabolic partial differential equation of the bidomain equations.

    If a stimulus is provided via

      cell = CardiacCellModel()
      cell.stimulus = Expression("I_s(t)", degree=1)

    then I_s is added to the right-hand side of (*), which thus reads

       d/dt v = - I(v, s) + I_s

    Note that the cardiac cell model stimulus is ignored when the cell
    model is used a spatially-varying setting (for instance in the
    bidomain setting). In this case, the user is expected to specify a
    stimulus for the cardiac model instead.
    """

    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict or  :py:class:`dolfin.Parameters`, optional)
           optional model parameters
         init_conditions (dict, optional)
           optional initial conditions
        """

        # FIXME: MER: Does this need to be this complicated?
        self._parameters = self.default_parameters()
        self._initial_conditions = self.default_initial_conditions()

        params = params or OrderedDict()
        init_conditions = init_conditions or OrderedDict()

        if params:
            assert isinstance(
                params, dict
            ), "expected a dict or a Parameters, as the params argument"
            if isinstance(params, dolfin.Parameters):
                params = params.to_dict()
            self.set_parameters(**params)

        if init_conditions:
            assert isinstance(
                init_conditions, dict
            ), "expected a dict or a Parameters, as the init_condition argument"
            if isinstance(init_conditions, dolfin.Parameters):
                init_conditions = init_conditions.to_dict()
            self.set_initial_conditions(**init_conditions)

        self.stimulus = None

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        return OrderedDict()

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        return OrderedDict()

    def set_parameters(self, **params):
        "Update parameters in model"
        for param_name, param_value in params.items():
            if param_name not in self._parameters:
                error("'%s' is not a parameter in %s" % (param_name, self))
            if not isinstance(param_value, (float, int)) and not isinstance(
                param_value._cpp_object, dolfin.cpp.function.GenericFunction
            ):
                error(
                    "'%s' is not a scalar or a dolfin.cpp.function.GenericFunction"
                    % param_name
                )
                if (
                    hasattr(param_value, "_cpp_object")
                    and isinstance(
                        param_value._cpp_object, dolfin.cpp.function.GenericFunction
                    )
                    and param_value._cpp_object.value_size() != 1
                ):
                    error("expected the value_size of '%s' to be 1" % param_name)

            self._parameters[param_name] = param_value

    def set_initial_conditions(self, **init):
        "Update initial_conditions in model"
        for init_name, init_value in init.items():
            if init_name not in self._initial_conditions:
                error("'%s' is not a parameter in %s" % (init_name, self))
            if not isinstance(init_value, (float, int)) and not isinstance(
                init_value._cpp_object, dolfin.cpp.function.GenericFunction
            ):
                error(
                    "'%s' is not a scalar or a dolfin.cpp.function.GenericFunction"
                    % init_name
                )
            if (
                hasattr(init_value, "_cpp_object")
                and isinstance(
                    init_value._cpp_object, dolfin.cpp.function.GenericFunction
                )
                and init_value._cpp_object.value_size() != 1
            ):
                error("expected the value_size of '%s' to be 1" % init_name)
            self._initial_conditions[init_name] = init_value

    def initial_conditions(self):
        "Return initial conditions for v and s as an Expression."
        return dolfin.Expression(
            list(self._initial_conditions.keys()), degree=1, **self._initial_conditions
        )

    def parameters(self):
        "Return the current parameters."
        return self._parameters

    def F(self, v, s, time=None):
        "Return right-hand side for state variable evolution."
        error("Must define F = F(v, s)")

    def I(self, v, s, time=None):
        "Return the ionic current."
        error("Must define I = I(v, s)")

    def num_states(self):
        """Return number of state variables (in addition to the
        membrane potential)."""
        error("Must overload num_states")

    def __str__(self):
        "Return string representation of class."
        return "Some cardiac cell model"


class MultiCellModel(CardiacCellModel):
    """
    MultiCellModel
    """

    def __init__(self, models, keys, markers):
        """
        *Arguments*
        models (tuple)
          tuple of existing CardiacCellModels
        keys (tuple)
          integers demarking the domain for each cell model
        markers (:py:class:`dolfin.MeshFunction`)
          MeshFunction defining the partitioning of the mesh (which model where)
        """
        self._cell_models = models
        self._keys = keys
        self._key_to_cell_model = dict(zip(keys, range(len(keys))))
        self._markers = markers

        self._num_states = max(c.num_states() for c in self._cell_models)

    def __getitem__(self, key):
        return self.models()[self._key_to_cell_model[key]]

    def models(self):
        return self._cell_models

    def markers(self):
        return self._markers

    def keys(self):
        return self._keys

    def mesh(self):
        return self._markers.mesh()

    def num_models(self):
        return len(self._cell_models)

    def num_states(self):
        """Return number of state variables (in addition to the
        membrane potential)."""
        return self._num_states

    def F(self, v, s, time=None, index=None):
        if index is None:
            error("(Domain) index must be specified for multi cell models")
        # Extract which cell model index (given by index in incoming tuple)
        k = self._key_to_cell_model[index]
        return self._cell_models[k].F(v, s, time)

    def I(self, v, s, time=None, index=None):
        if index is None:
            error("(Domain) index must be specified for multi cell models")
        # Extract which cell model index (given by index in incoming tuple)
        k = self._key_to_cell_model[index]
        return self._cell_models[k].I(v, s, time)

    def initial_conditions(self):
        "Return initial conditions for v and s as a dolfin.dolfin.cpp.function.GenericFunction."

        n = self.num_states()  # (Maximal) Number of states in MultiCellModel
        VS = dolfin.VectorFunctionSpace(self.mesh(), "DG", 0, n + 1)
        vs = backend.Function(VS)

        markers = self.markers()
        u = dolfin.TrialFunction(VS)
        v = dolfin.TestFunction(VS)

        dy = dolfin.Measure("dx", domain=self.mesh(), subdomain_data=markers)

        # Define projection into multiverse
        a = dolfin.inner(u, v) * dy()

        Ls = list()
        for k, model in enumerate(self.models()):
            ic = model.initial_conditions()  # Extract initial conditions
            n_k = model.num_states()  # Extract number of local states
            i_k = self.keys()[k]  # Extract domain index of cell model k
            L_k = sum(ic[j] * v[j] * dy(i_k) for j in range(n_k))
            Ls.append(L_k)
        L = sum(Ls)
        backend.solve(a == L, vs)
        return vs


# from dolfin import *
# from cbcbeat import *

# mesh = UnitSquareMesh(20, 20)
# markers = CellFunction("uint", mesh, 0)
# markers.array()[::2] = 1
# c0 = Beeler_reuter_1977()
# c1 = FitzHughNagumoManual()
# cell_model = MultiCellModel((c0, c1), (0, 1), markers)
# print cell_model.num_states()



"""This module contains a CellSolver that uses JIT compiled Gotran
models together with GOSS (General ODE System Solver), which can be
interfaced by the GossSplittingSolver"""

__author__ = "Johan Hake (hake.dev@gmail.com), 2013"

__all__ = ["GOSSplittingSolver"]

from cbcbeat.cellmodels.cardiaccellmodel import MultiCellModel

from dolfin import Parameters

from goss.dolfinutils import DOLFINODESystemSolver

# cbcbeat imports
from cbcbeat.bidomainsolver import BidomainSolver
from cbcbeat.monodomainsolver import MonodomainSolver
from cbcbeat.splittingsolver import SplittingSolver
from cbcbeat.utils import Projecter


class GOSSplittingSolver(SplittingSolver):
    @staticmethod
    def default_parameters():
        """Initialize and return a set of default parameters for the
        splitting solver

        *Returns*
          A set of parameters (:py:class:`dolfin.Parameters`)

        To inspect all the default parameters, do::

          info(SplittingSolver.default_parameters(), True)
        """

        params = Parameters("GOSSplittingSolver")

        # Have to be false as GOSS is not compatible with dolfin-adjoint
        params.add("apply_stimulus_current_to_pde", False)
        params.add("enable_adjoint", False)
        params.add("theta", 0.5, 0, 1)
        try:
            params.add("pde_solver", "bidomain", set(["bidomain", "monodomain"]))
        except Exception:
            params.add("pde_solver", "bidomain", ["bidomain", "monodomain"])
            pass

        # Add default parameters from ODE solver
        ode_solver_params = DOLFINODESystemSolver.default_parameters_dolfin()
        ode_solver_params.rename("ode_solver")
        # ode_solver_params.add("membrane_potential", "V")
        params.add(ode_solver_params)

        pde_solver_params = BidomainSolver.default_parameters()
        pde_solver_params["polynomial_degree"] = 1
        params.add(pde_solver_params)

        pde_solver_params = MonodomainSolver.default_parameters()
        pde_solver_params["polynomial_degree"] = 1
        params.add(pde_solver_params)

        projecter_params = Projecter.default_parameters()
        params.add(projecter_params)

        return params

    def _create_ode_solver(self):
        """Helper function to initialize a suitable ODE solver from
        the cardiac model."""

        # Extract cardiac cell model from cardiac model
        cell_models = self._model.cell_models()

        kwargs = {"odes": cell_models}
        if isinstance(cell_models, MultiCellModel):
            kwargs = {
                "odes": dict(zip(cell_models.keys(), cell_models.models())),
                "domains": cell_models.markers(),
            }

        solver = DOLFINODESystemSolver(
            self._domain, params=self.parameters["ode_solver"], **kwargs
        )
        return solver



"""This module handles all dolfin import in cbcbeat. Here dolfin and
dolfin_adjoint gets imported. If dolfin_adjoint is not present it will not
be imported."""

__author__ = "Johan Hake (hake.dev@gmail.com), 2013"


import dolfin as backend

try:
    import dolfin_adjoint as backend  # noqa: F811

    has_dolfin_adjoint = True

except ImportError:
    # FIXME: Should we raise some sort of warning?
    has_dolfin_adjoint = False


__all__ = ["backend"]



"""This module contains a Rogers-McCulloch cardiac cell model which is
a modified version of the FitzHughNagumo model.

This formulation is based on the description on page 2 of "Optimal
control approach ..." by Nagaiah, Kunisch and Plank, 2013, J Math
Biol.

The module was written by hand, in particular it was not
autogenerated.

"""

from __future__ import division

__author__ = "Marie E. Rognes (meg@simula.no), 2012--2013"
__all__ = ["RogersMcCulloch"]

from collections import OrderedDict
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class RogersMcCulloch(CardiacCellModel):
    """The Rogers-McCulloch model is a modified FitzHughNagumo model. This
    formulation follows the description on page 2 of "Optimal control
    approach ..." by Nagaiah, Kunisch and Plank, 2013, J Math Biol
    with w replaced by s. Note that this model introduces one
    additional parameter compared to the original 1994
    Rogers-McCulloch model.

    This is a model containing two nonlinear, ODEs for the evolution
    of the transmembrane potential v and one additional state variable
    s:

    .. math::

      \frac{dv}{dt} = - I_{ion}(v, s)

      \frac{ds}{dt} = F(v, s)

    where

    .. math::

      I_{ion}(v, s) = g v (1 - v/v_th)(1 - v/v_p) + \eta_1 v s

          F(v, s) = \eta_2  (v/vp - \eta_3 s)

    """

    def __init__(self, params=None, init_conditions=None):
        "Create cardiac cell model, optionally from given parameters."
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("g", 1.5),  # S/(cm^2)
                ("v_th", 13.0),  # mV
                ("v_p", 100.0),  # mV
                ("eta_1", 4.4),  # S/(cm^2)
                ("eta_2", 0.012),
                ("eta_3", 1.0),
            ]
        )
        # The original parameters of Rogers and McCulloch, 1994.
        # params = OrderedDict([("a", 0.13),
        #                      ("b", 0.013),
        #                      ("d", 1.),
        #                      ("c_1", 0.26),
        #                      ("c_2", 0.1)])

        return params

    def I(self, v, s, time=None):
        "Return the ionic current."

        # Extract parameters
        g = self._parameters["g"]
        v_p = self._parameters["v_p"]
        v_th = self._parameters["v_th"]
        eta_1 = self._parameters["eta_1"]
        # Define current
        i = g * v * (1 - v / v_th) * (1 - v / v_p) + eta_1 * v * s

        # Original R&McC current
        # a = self._parameters["a"]
        # c_1 = self._parameters["c_1"]
        # c_2 = self._parameters["c_2"]
        # i = - (c_1*v*(v - a)*(1 - v) - c_2*v*s)

        return i

    def F(self, v, s, time=None):
        "Return right-hand side for state variable evolution."

        # Extract parameters
        eta_2 = self._parameters["eta_2"]
        eta_3 = self._parameters["eta_3"]
        v_p = self._parameters["v_p"]
        f = eta_2 * (v / v_p - eta_3 * s)

        # The original from R&McC 1994
        # b = self._parameters["b"]
        # d = self._parameters["d"]
        # f = b*(v - d*s)
        return f

    @staticmethod
    def default_initial_conditions():
        ic = OrderedDict([("V", 0.0), ("S", 0.0)])
        return ic

    def num_states(self):
        "Return number of state variables."
        return 1

    def __str__(self):
        "Return string representation of class."
        return "Rogers-McCulloch 1994 cardiac cell model"



"""This module contains a Tentusscher_2004_mcell cardiac cell model

The module was autogenerated from a gotran ode file
"""
from __future__ import division
from collections import OrderedDict
import ufl
import dolfin

from cbcbeat.dolfinimport import backend
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class Tentusscher_2004_mcell(CardiacCellModel):
    """
    NOT_IMPLEMENTED
    """

    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        """
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("g_Kr", 0.096),
                ("g_Ks", 0.062),
                ("g_Na", 14.838),
                ("g_CaL", 0.000175),
                ("g_to", 0.294),
                ("Buf_c", 0.15),
                ("Buf_sr", 10),
                ("Ca_o", 2),
                ("K_buf_c", 0.001),
                ("K_buf_sr", 0.3),
                ("K_up", 0.00025),
                ("V_leak", 8e-05),
                ("V_sr", 0.001094),
                ("Vmax_up", 0.000425),
                ("a_rel", 0.016464),
                ("b_rel", 0.25),
                ("c_rel", 0.008232),
                ("tau_g", 2),
                ("Na_o", 140),
                ("Cm", 0.185),
                ("F", 96485.3415),
                ("R", 8314.472),
                ("T", 310),
                ("V_c", 0.016404),
                ("stim_amplitude", 0),
                ("stim_duration", 1),
                ("stim_period", 500),
                ("stim_start", 1),
                ("K_o", 5.4),
                ("P_kna", 0.03),
                ("g_K1", 5.405),
                ("g_bna", 0.00029),
                ("g_bca", 0.000592),
                ("K_mNa", 40),
                ("K_mk", 1),
                ("P_NaK", 1.362),
                ("K_NaCa", 1000),
                ("K_sat", 0.1),
                ("Km_Ca", 1.38),
                ("Km_Nai", 87.5),
                ("alpha", 2.5),
                ("gamma", 0.35),
                ("K_pCa", 0.0005),
                ("g_pCa", 0.825),
                ("g_pK", 0.0146),
            ]
        )
        return params

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict(
            [
                ("V", -86.2),
                ("Xr1", 0.0),
                ("Xr2", 1.0),
                ("Xs", 0.0),
                ("m", 0.0),
                ("h", 0.75),
                ("j", 0.75),
                ("d", 0.0),
                ("f", 1.0),
                ("fCa", 1.0),
                ("s", 1.0),
                ("r", 0.0),
                ("Ca_SR", 0.2),
                ("Ca_i", 0.0002),
                ("g", 1.0),
                ("Na_i", 11.6),
                ("K_i", 138.3),
            ]
        )
        return ic

    def _I(self, v, s, time):
        """
        Original gotran transmembrane current dV/dt
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 16
        Xr1, Xr2, Xs, m, h, j, d, f, fCa, s, r, Ca_SR, Ca_i, g, Na_i, K_i = s

        # Assign parameters
        P_kna = self._parameters["P_kna"]
        g_K1 = self._parameters["g_K1"]
        g_Kr = self._parameters["g_Kr"]
        g_Ks = self._parameters["g_Ks"]
        g_Na = self._parameters["g_Na"]
        g_bna = self._parameters["g_bna"]
        g_CaL = self._parameters["g_CaL"]
        g_bca = self._parameters["g_bca"]
        g_to = self._parameters["g_to"]
        K_mNa = self._parameters["K_mNa"]
        K_mk = self._parameters["K_mk"]
        P_NaK = self._parameters["P_NaK"]
        K_NaCa = self._parameters["K_NaCa"]
        K_sat = self._parameters["K_sat"]
        Km_Ca = self._parameters["Km_Ca"]
        Km_Nai = self._parameters["Km_Nai"]
        alpha = self._parameters["alpha"]
        gamma = self._parameters["gamma"]
        K_pCa = self._parameters["K_pCa"]
        g_pCa = self._parameters["g_pCa"]
        g_pK = self._parameters["g_pK"]
        Ca_o = self._parameters["Ca_o"]
        Na_o = self._parameters["Na_o"]
        F = self._parameters["F"]
        R = self._parameters["R"]
        T = self._parameters["T"]
        stim_amplitude = self._parameters["stim_amplitude"]
        stim_duration = self._parameters["stim_duration"]
        stim_start = self._parameters["stim_start"]
        K_o = self._parameters["K_o"]

        # Init return args
        current = [ufl.zero()] * 1

        # Expressions for the Reversal potentials component
        E_Na = R * T * ufl.ln(Na_o / Na_i) / F
        E_K = R * T * ufl.ln(K_o / K_i) / F
        E_Ks = R * T * ufl.ln((Na_o * P_kna + K_o) / (K_i + P_kna * Na_i)) / F
        E_Ca = 0.5 * R * T * ufl.ln(Ca_o / Ca_i) / F

        # Expressions for the Inward rectifier potassium current component
        alpha_K1 = 0.1 / (1 + 6.14421235333e-06 * ufl.exp(-0.06 * E_K + 0.06 * V))
        beta_K1 = (
            3.06060402008 * ufl.exp(0.0002 * V - 0.0002 * E_K)
            + 0.367879441171 * ufl.exp(0.1 * V - 0.1 * E_K)
        ) / (1 + ufl.exp(0.5 * E_K - 0.5 * V))
        xK1_inf = alpha_K1 / (alpha_K1 + beta_K1)
        i_K1 = 0.430331482912 * g_K1 * ufl.sqrt(K_o) * (-E_K + V) * xK1_inf

        # Expressions for the Rapid time dependent potassium current component
        i_Kr = 0.430331482912 * g_Kr * ufl.sqrt(K_o) * (-E_K + V) * Xr1 * Xr2

        # Expressions for the Slow time dependent potassium current component
        i_Ks = g_Ks * (Xs * Xs) * (-E_Ks + V)

        # Expressions for the Fast sodium current component
        i_Na = g_Na * (m * m * m) * (-E_Na + V) * h * j

        # Expressions for the Sodium background current component
        i_b_Na = g_bna * (-E_Na + V)

        # Expressions for the L type ca current component
        i_CaL = (
            4
            * g_CaL
            * (F * F)
            * (Ca_i * ufl.exp(2 * F * V / (R * T)) - 0.341 * Ca_o)
            * V
            * d
            * f
            * fCa
            / (R * T * (-1 + ufl.exp(2 * F * V / (R * T))))
        )

        # Expressions for the Calcium background current component
        i_b_Ca = g_bca * (-E_Ca + V)

        # Expressions for the Transient outward current component
        i_to = g_to * (-E_K + V) * r * s

        # Expressions for the Sodium potassium pump current component
        i_NaK = (
            K_o
            * P_NaK
            * Na_i
            / (
                (K_mNa + Na_i)
                * (K_mk + K_o)
                * (
                    1
                    + 0.0353 * ufl.exp(-F * V / (R * T))
                    + 0.1245 * ufl.exp(-0.1 * F * V / (R * T))
                )
            )
        )

        # Expressions for the Sodium calcium exchanger current component
        i_NaCa = (
            K_NaCa
            * (
                -alpha
                * (Na_o * Na_o * Na_o)
                * Ca_i
                * ufl.exp(F * (-1 + gamma) * V / (R * T))
                + Ca_o * (Na_i * Na_i * Na_i) * ufl.exp(F * gamma * V / (R * T))
            )
            / (
                (1 + K_sat * ufl.exp(F * (-1 + gamma) * V / (R * T)))
                * (Km_Ca + Ca_o)
                * ((Na_o * Na_o * Na_o) + (Km_Nai * Km_Nai * Km_Nai))
            )
        )

        # Expressions for the Calcium pump current component
        i_p_Ca = g_pCa * Ca_i / (Ca_i + K_pCa)

        # Expressions for the Potassium pump current component
        i_p_K = g_pK * (-E_K + V) / (1 + 65.4052157419 * ufl.exp(-0.167224080268 * V))

        # Expressions for the Membrane component
        i_Stim = (
            -stim_amplitude
            * (1 - 1 / (1 + ufl.exp(5.0 * time - 5.0 * stim_start)))
            / (1 + ufl.exp(5.0 * time - 5.0 * stim_start - 5.0 * stim_duration))
        )
        current[0] = (
            -i_Ks
            - i_NaCa
            - i_CaL
            - i_b_Na
            - i_Kr
            - i_p_Ca
            - i_to
            - i_b_Ca
            - i_Stim
            - i_Na
            - i_p_K
            - i_NaK
            - i_K1
        )

        # Return results
        return current[0]

    def I(self, v, s, time=None):
        """
        Transmembrane current

           I = -dV/dt

        """
        return -self._I(v, s, time)

    def F(self, v, s, time=None):
        """
        Right hand side for ODE system
        """
        time = time if time else backend.Constant(0.0)

        # Assign states
        V = v
        assert len(s) == 16
        Xr1, Xr2, Xs, m, h, j, d, f, fCa, s, r, Ca_SR, Ca_i, g, Na_i, K_i = s

        # Assign parameters
        P_kna = self._parameters["P_kna"]
        g_K1 = self._parameters["g_K1"]
        g_Kr = self._parameters["g_Kr"]
        g_Ks = self._parameters["g_Ks"]
        g_Na = self._parameters["g_Na"]
        g_bna = self._parameters["g_bna"]
        g_CaL = self._parameters["g_CaL"]
        g_bca = self._parameters["g_bca"]
        g_to = self._parameters["g_to"]
        K_mNa = self._parameters["K_mNa"]
        K_mk = self._parameters["K_mk"]
        P_NaK = self._parameters["P_NaK"]
        K_NaCa = self._parameters["K_NaCa"]
        K_sat = self._parameters["K_sat"]
        Km_Ca = self._parameters["Km_Ca"]
        Km_Nai = self._parameters["Km_Nai"]
        alpha = self._parameters["alpha"]
        gamma = self._parameters["gamma"]
        K_pCa = self._parameters["K_pCa"]
        g_pCa = self._parameters["g_pCa"]
        g_pK = self._parameters["g_pK"]
        Buf_c = self._parameters["Buf_c"]
        Buf_sr = self._parameters["Buf_sr"]
        Ca_o = self._parameters["Ca_o"]
        K_buf_c = self._parameters["K_buf_c"]
        K_buf_sr = self._parameters["K_buf_sr"]
        K_up = self._parameters["K_up"]
        V_leak = self._parameters["V_leak"]
        V_sr = self._parameters["V_sr"]
        Vmax_up = self._parameters["Vmax_up"]
        a_rel = self._parameters["a_rel"]
        b_rel = self._parameters["b_rel"]
        c_rel = self._parameters["c_rel"]
        tau_g = self._parameters["tau_g"]
        Na_o = self._parameters["Na_o"]
        Cm = self._parameters["Cm"]
        F = self._parameters["F"]
        R = self._parameters["R"]
        T = self._parameters["T"]
        V_c = self._parameters["V_c"]
        stim_amplitude = self._parameters["stim_amplitude"]
        stim_duration = self._parameters["stim_duration"]
        stim_start = self._parameters["stim_start"]
        K_o = self._parameters["K_o"]

        # Init return args
        F_expressions = [ufl.zero()] * 16

        # Expressions for the Reversal potentials component
        E_Na = R * T * ufl.ln(Na_o / Na_i) / F
        E_K = R * T * ufl.ln(K_o / K_i) / F
        E_Ks = R * T * ufl.ln((Na_o * P_kna + K_o) / (K_i + P_kna * Na_i)) / F
        E_Ca = 0.5 * R * T * ufl.ln(Ca_o / Ca_i) / F

        # Expressions for the Inward rectifier potassium current component
        alpha_K1 = 0.1 / (1 + 6.14421235333e-06 * ufl.exp(-0.06 * E_K + 0.06 * V))
        beta_K1 = (
            3.06060402008 * ufl.exp(0.0002 * V - 0.0002 * E_K)
            + 0.367879441171 * ufl.exp(0.1 * V - 0.1 * E_K)
        ) / (1 + ufl.exp(0.5 * E_K - 0.5 * V))
        xK1_inf = alpha_K1 / (alpha_K1 + beta_K1)
        i_K1 = 0.430331482912 * g_K1 * ufl.sqrt(K_o) * (-E_K + V) * xK1_inf

        # Expressions for the Rapid time dependent potassium current component
        i_Kr = 0.430331482912 * g_Kr * ufl.sqrt(K_o) * (-E_K + V) * Xr1 * Xr2

        # Expressions for the xr1 gate component
        xr1_inf = 1.0 / (1 + 0.0243728440733 * ufl.exp(-0.142857142857 * V))
        alpha_xr1 = 450.0 / (1 + ufl.exp(-9 / 2 - V / 10))
        beta_xr1 = 6.0 / (1 + 13.5813245226 * ufl.exp(0.0869565217391 * V))
        tau_xr1 = alpha_xr1 * beta_xr1
        F_expressions[0] = (xr1_inf - Xr1) / tau_xr1

        # Expressions for the xr2 gate component
        xr2_inf = 1.0 / (1 + 39.1212839982 * ufl.exp(0.0416666666667 * V))
        alpha_xr2 = 3 / (1 + 0.0497870683679 * ufl.exp(-0.05 * V))
        beta_xr2 = 1.12 / (1 + 0.0497870683679 * ufl.exp(0.05 * V))
        tau_xr2 = alpha_xr2 * beta_xr2
        F_expressions[1] = (xr2_inf - Xr2) / tau_xr2

        # Expressions for the Slow time dependent potassium current component
        i_Ks = g_Ks * (Xs * Xs) * (-E_Ks + V)

        # Expressions for the xs gate component
        xs_inf = 1.0 / (1 + 0.699672537375 * ufl.exp(-0.0714285714286 * V))
        alpha_xs = 1100.0 / ufl.sqrt(1 + 0.188875602838 * ufl.exp(-0.166666666667 * V))
        beta_xs = 1.0 / (1 + 0.0497870683679 * ufl.exp(0.05 * V))
        tau_xs = alpha_xs * beta_xs
        F_expressions[2] = (xs_inf - Xs) / tau_xs

        # Expressions for the Fast sodium current component
        i_Na = g_Na * (m * m * m) * (-E_Na + V) * h * j

        # Expressions for the m gate component
        m_inf = 1.0 / (
            (1 + 0.00184221158117 * ufl.exp(-0.110741971207 * V))
            * (1 + 0.00184221158117 * ufl.exp(-0.110741971207 * V))
        )
        alpha_m = 1.0 / (1 + ufl.exp(-12 - V / 5))
        beta_m = 0.1 / (1 + ufl.exp(7 + V / 5)) + 0.1 / (
            1 + 0.778800783071 * ufl.exp(0.005 * V)
        )
        tau_m = alpha_m * beta_m
        F_expressions[3] = (-m + m_inf) / tau_m

        # Expressions for the h gate component
        h_inf = 1.0 / (
            (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
            * (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
        )
        alpha_h = (
            4.43126792958e-07
            * ufl.exp(-0.147058823529 * V)
            / (1 + 2.35385266837e17 * ufl.exp(1.0 * V))
        )
        beta_h = (2.7 * ufl.exp(0.079 * V) + 310000 * ufl.exp(0.3485 * V)) / (
            1 + 2.35385266837e17 * ufl.exp(1.0 * V)
        ) + 0.77 * (1 - 1 / (1 + 2.35385266837e17 * ufl.exp(1.0 * V))) / (
            0.13 + 0.0497581410839 * ufl.exp(-0.0900900900901 * V)
        )
        tau_h = 1.0 / (alpha_h + beta_h)
        F_expressions[4] = (-h + h_inf) / tau_h

        # Expressions for the j gate component
        j_inf = 1.0 / (
            (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
            * (1 + 15212.5932857 * ufl.exp(0.134589502019 * V))
        )
        alpha_j = (
            (37.78 + V)
            * (-25428 * ufl.exp(0.2444 * V) - 6.948e-06 * ufl.exp(-0.04391 * V))
            / (
                (1 + 2.35385266837e17 * ufl.exp(1.0 * V))
                * (1 + 50262745826.0 * ufl.exp(0.311 * V))
            )
        )
        beta_j = 0.6 * (1 - 1 / (1 + 2.35385266837e17 * ufl.exp(1.0 * V))) * ufl.exp(
            0.057 * V
        ) / (1 + 0.0407622039784 * ufl.exp(-0.1 * V)) + 0.02424 * ufl.exp(
            -0.01052 * V
        ) / (
            (1 + 2.35385266837e17 * ufl.exp(1.0 * V))
            * (1 + 0.0039608683399 * ufl.exp(-0.1378 * V))
        )
        tau_j = 1.0 / (alpha_j + beta_j)
        F_expressions[5] = (-j + j_inf) / tau_j

        # Expressions for the Sodium background current component
        i_b_Na = g_bna * (-E_Na + V)

        # Expressions for the L type ca current component
        i_CaL = (
            4
            * g_CaL
            * (F * F)
            * (Ca_i * ufl.exp(2 * F * V / (R * T)) - 0.341 * Ca_o)
            * V
            * d
            * f
            * fCa
            / (R * T * (-1 + ufl.exp(2 * F * V / (R * T))))
        )

        # Expressions for the d gate component
        d_inf = 1.0 / (1 + 0.513417119033 * ufl.exp(-0.133333333333 * V))
        alpha_d = 0.25 + 1.4 / (1 + 0.0677244716592 * ufl.exp(-0.0769230769231 * V))
        beta_d = 1.4 / (1 + ufl.exp(1 + V / 5))
        gamma_d = 1.0 / (1 + 12.1824939607 * ufl.exp(-0.05 * V))
        tau_d = alpha_d * beta_d + gamma_d
        F_expressions[6] = (-d + d_inf) / tau_d

        # Expressions for the f gate component
        f_inf = 1.0 / (1 + 17.4117080633 * ufl.exp(0.142857142857 * V))
        tau_f = (
            80
            + 165.0 / (1 + ufl.exp(5 / 2 - V / 10))
            + 1125 * ufl.exp(-0.00416666666667 * ((27 + V) * (27 + V)))
        )
        F_expressions[7] = (f_inf - f) / tau_f

        # Expressions for the fca gate component
        alpha_fCa = 1.0 / (1 + 8.03402376702e27 * ufl.elem_pow(Ca_i, 8))
        beta_fCa = 0.1 / (1 + 0.00673794699909 * ufl.exp(10000.0 * Ca_i))
        gama_fCa = 0.2 / (1 + 0.391605626677 * ufl.exp(1250.0 * Ca_i))
        fCa_inf = (
            0.157534246575
            + 0.684931506849 * gama_fCa
            + 0.684931506849 * alpha_fCa
            + 0.684931506849 * beta_fCa
        )
        tau_fCa = 2
        d_fCa = (-fCa + fCa_inf) / tau_fCa
        F_expressions[8] = (
            1
            - 1.0 / ((1 + ufl.exp(60 + V)) * (1 + ufl.exp(10.0 * fCa_inf - 10.0 * fCa)))
        ) * d_fCa

        # Expressions for the Calcium background current component
        i_b_Ca = g_bca * (-E_Ca + V)

        # Expressions for the Transient outward current component
        i_to = g_to * (-E_K + V) * r * s

        # Expressions for the s gate component
        s_inf = 1.0 / (1 + ufl.exp(4 + V / 5))
        tau_s = (
            3
            + 5 / (1 + ufl.exp(-4 + V / 5))
            + 85 * ufl.exp(-0.003125 * ((45 + V) * (45 + V)))
        )
        F_expressions[9] = (s_inf - s) / tau_s

        # Expressions for the r gate component
        r_inf = 1.0 / (1 + 28.0316248945 * ufl.exp(-0.166666666667 * V))
        tau_r = 0.8 + 9.5 * ufl.exp(-0.000555555555556 * ((40 + V) * (40 + V)))
        F_expressions[10] = (r_inf - r) / tau_r

        # Expressions for the Sodium potassium pump current component
        i_NaK = (
            K_o
            * P_NaK
            * Na_i
            / (
                (K_mNa + Na_i)
                * (K_mk + K_o)
                * (
                    1
                    + 0.0353 * ufl.exp(-F * V / (R * T))
                    + 0.1245 * ufl.exp(-0.1 * F * V / (R * T))
                )
            )
        )

        # Expressions for the Sodium calcium exchanger current component
        i_NaCa = (
            K_NaCa
            * (
                -alpha
                * (Na_o * Na_o * Na_o)
                * Ca_i
                * ufl.exp(F * (-1 + gamma) * V / (R * T))
                + Ca_o * (Na_i * Na_i * Na_i) * ufl.exp(F * gamma * V / (R * T))
            )
            / (
                (1 + K_sat * ufl.exp(F * (-1 + gamma) * V / (R * T)))
                * (Km_Ca + Ca_o)
                * ((Na_o * Na_o * Na_o) + (Km_Nai * Km_Nai * Km_Nai))
            )
        )

        # Expressions for the Calcium pump current component
        i_p_Ca = g_pCa * Ca_i / (Ca_i + K_pCa)

        # Expressions for the Potassium pump current component
        i_p_K = g_pK * (-E_K + V) / (1 + 65.4052157419 * ufl.exp(-0.167224080268 * V))

        # Expressions for the Calcium dynamics component
        i_rel = (
            (c_rel + a_rel * (Ca_SR * Ca_SR) / ((b_rel * b_rel) + (Ca_SR * Ca_SR)))
            * d
            * g
        )
        i_up = Vmax_up / (1 + (K_up * K_up) / (Ca_i * Ca_i))
        i_leak = V_leak * (Ca_SR - Ca_i)
        g_inf = 1 / (
            (1 + 0.0301973834223 * ufl.exp(10000.0 * Ca_i))
            * (1 + 5.43991024148e20 * ufl.elem_pow(Ca_i, 6))
        ) + (1 - 1 / (1 + 0.0301973834223 * ufl.exp(10000.0 * Ca_i))) / (
            1 + 1.9720198874e55 * ufl.elem_pow(Ca_i, 16)
        )
        d_g = (-g + g_inf) / tau_g
        Ca_i_bufc = 1.0 / (1 + Buf_c * K_buf_c / ((Ca_i + K_buf_c) * (Ca_i + K_buf_c)))
        Ca_sr_bufsr = 1.0 / (
            1 + Buf_sr * K_buf_sr / ((Ca_SR + K_buf_sr) * (Ca_SR + K_buf_sr))
        )
        F_expressions[13] = (
            1 - 1.0 / ((1 + ufl.exp(60 + V)) * (1 + ufl.exp(10.0 * g_inf - 10.0 * g)))
        ) * d_g
        F_expressions[12] = (
            i_rel
            - i_up
            - Cm * (i_b_Ca + i_p_Ca - 2 * i_NaCa + i_CaL) / (2 * F * V_c)
            + i_leak
        ) * Ca_i_bufc
        F_expressions[11] = V_c * (i_up - i_leak - i_rel) * Ca_sr_bufsr / V_sr

        # Expressions for the Sodium dynamics component
        F_expressions[14] = Cm * (-i_b_Na - i_Na - 3 * i_NaCa - 3 * i_NaK) / (F * V_c)

        # Expressions for the Membrane component
        i_Stim = (
            -stim_amplitude
            * (1 - 1 / (1 + ufl.exp(5.0 * time - 5.0 * stim_start)))
            / (1 + ufl.exp(5.0 * time - 5.0 * stim_start - 5.0 * stim_duration))
        )

        # Expressions for the Potassium dynamics component
        F_expressions[15] = (
            Cm * (2 * i_NaK - i_Ks - i_Kr - i_to - i_Stim - i_p_K - i_K1) / (F * V_c)
        )

        # Return results
        return dolfin.as_vector(F_expressions)

    def num_states(self):
        return 16

    def __str__(self):
        return "Tentusscher_2004_mcell cardiac cell model"



"""This module contains a FitzHugh-Nagumo cardiac cell model

The module was written by hand, in particular it was not
autogenerated.
"""

from __future__ import division

__author__ = "Marie E. Rognes (meg@simula.no), 2012--2013"
__all__ = ["FitzHughNagumoManual"]

from collections import OrderedDict
from cbcbeat.cellmodels.cardiaccellmodel import CardiacCellModel


class FitzHughNagumoManual(CardiacCellModel):
    """
    A reparametrized FitzHughNagumo model, based on Section 2.4.1 in
    "Computing the electrical activity in the heart" by Sundnes et al,
    2006.

    This is a model containing two nonlinear, ODEs for the evolution
    of the transmembrane potential v and one additional state variable
    s.
    """

    def __init__(self, params=None, init_conditions=None):
        "Create cardiac cell model, optionally from given parameters."
        CardiacCellModel.__init__(self, params, init_conditions)

    @staticmethod
    def default_parameters():
        "Set-up and return default parameters."
        params = OrderedDict(
            [
                ("a", 0.13),
                ("b", 0.013),
                ("c_1", 0.26),
                ("c_2", 0.1),
                ("c_3", 1.0),
                ("v_peak", 40.0),
                ("v_rest", -85.0),
            ]
        )
        return params

    def I(self, v, s, time=None):
        "Return the ionic current."
        # Extract parameters
        c_1 = self._parameters["c_1"]
        c_2 = self._parameters["c_2"]
        v_rest = self._parameters["v_rest"]
        v_peak = self._parameters["v_peak"]
        v_amp = v_peak - v_rest
        v_th = v_rest + self._parameters["a"] * v_amp

        # Define current
        i = (
            c_1 / (v_amp**2) * (v - v_rest) * (v - v_th) * (v_peak - v)
            - c_2 / (v_amp) * (v - v_rest) * s
        )
        return -i

    def F(self, v, s, time=None):
        "Return right-hand side for state variable evolution."

        # Extract parameters
        b = self._parameters["b"]
        v_rest = self._parameters["v_rest"]
        c_3 = self._parameters["c_3"]

        # Define model
        return b * (v - v_rest - c_3 * s)

    @staticmethod
    def default_initial_conditions():
        ic = OrderedDict([("V", -85.0), ("S", 0.0)])
        return ic

    def num_states(self):
        "Return number of state variables."
        return 1

    def __str__(self):
        "Return string representation of class."
        return "(Manual) FitzHugh-Nagumo cardiac cell model"
